# pylint: disable=missing-docstring
import os
import uuid

from resolwe.flow.models import Data, Secret
from resolwe.test import ProcessTestCase, TransactionTestCase, tag_process, with_docker_executor

PROCESSES_DIR = os.path.join(os.path.dirname(__file__), 'processes')


class SecretsModelTest(TransactionTestCase):
    def test_secret_manager(self):
        handle = Secret.objects.create_secret('foo bar', self.user)
        value = Secret.objects.get_secret(handle, self.user)
        self.assertEqual(value, 'foo bar')

        with self.assertRaises(Secret.DoesNotExist):
            value = Secret.objects.get_secret(handle, self.admin)

        with self.assertRaises(Secret.DoesNotExist):
            value = Secret.objects.get_secret(uuid.uuid4(), self.user)


class SecretsProcessTest(ProcessTestCase):
    def setUp(self):
        super().setUp()

        self._register_schemas(processes_paths=[PROCESSES_DIR])

        # Create some secrets.
        self.secret_value = 'hello secret world'
        self.secret_handle = Secret.objects.create_secret(self.secret_value, self.admin)

    @with_docker_executor
    @tag_process('test-secrets-permission-denied')
    def test_permission_denied(self):
        data = self.run_process(
            'test-secrets-permission-denied',
            input_={'token': {'handle': self.secret_handle}},
            assert_status=Data.STATUS_ERROR
        )
        self.assertIn(
            "Permission denied for process: Process 'test-secrets-permission-denied' has secret "
            "inputs, but no permission to see secrets",
            data.process_error
        )

    @with_docker_executor
    @tag_process('test-secrets-echo')
    def test_secret_echo(self):
        data = self.run_process('test-secrets-echo', input_={'token': {'handle': self.secret_handle}})
        self.assertEqual(data.output['secret'], self.secret_value)

    @tag_process('test-secrets-echo')
    def test_secret_local_executor(self):
        data = self.run_process(
            'test-secrets-echo',
            input_={'token': {'handle': self.secret_handle}},
            assert_status=Data.STATUS_ERROR
        )
        self.assertIn(
            "Permission denied for process: Process which requires access to secrets cannot be run "
            "using the local executor",
            data.process_error
        )



""".. Ignore pydocstyle D400.

==================
Elastic Paginators
==================

Paginator classes used in Elastic app.


.. autoclass:: resolwe.elastic.pagination.LimitOffsetPostPagination

"""
from rest_framework.pagination import LimitOffsetPagination, _positive_int


def get_query_param(request, key):
    """Get query parameter uniformly for GET and POST requests."""
    value = request.query_params.get(key) or request.data.get(key)
    if value is None:
        raise KeyError()
    return value


class LimitOffsetPostPagination(LimitOffsetPagination):
    """Limit/offset paginator.

    This is standard limit/offset paginator from Django REST framework,
    with difference that it supports passing ``limit`` and ``offset``
    attributes also in the body of the request (not just as query
    parameter).
    """

    def get_limit(self, request):
        """Return limit parameter."""
        if self.limit_query_param:
            try:
                return _positive_int(
                    get_query_param(request, self.limit_query_param),
                    strict=True,
                    cutoff=self.max_limit
                )
            except (KeyError, ValueError):
                pass

        return self.default_limit

    def get_offset(self, request):
        """Return offset parameter."""
        try:
            return _positive_int(
                get_query_param(request, self.offset_query_param),
            )
        except (KeyError, ValueError):
            return 0



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-09-21 10:43
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('flow', '0003_data_dependency_1'),
    ]

    operations = [
        migrations.AddField(
            model_name='datadependency',
            name='kind',
            field=models.CharField(choices=[('io', 'Input/output dependency'), ('subprocess', 'Subprocess')], default='io', max_length=16),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='datadependency',
            name='child',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='parents_dependency', to='flow.Data'),
        ),
        migrations.AlterField(
            model_name='datadependency',
            name='parent',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='children_dependency', to='flow.Data'),
        ),
        migrations.AlterModelTable(
            name='datadependency',
            table=None,
        ),
    ]



"""Resolwe collection model."""
import datetime

from django.contrib.postgres.fields import ArrayField, JSONField
from django.db import models, transaction

from resolwe.permissions.shortcuts import get_objects_for_user
from resolwe.permissions.utils import assign_contributor_permissions

from .base import BaseModel
from .utils import DirtyError, validate_schema


class BaseCollection(BaseModel):
    """Template for Postgres model for storing a collection."""

    class Meta(BaseModel.Meta):
        """BaseCollection Meta options."""

        abstract = True

    #: detailed description
    description = models.TextField(blank=True)

    settings = JSONField(default=dict)

    data = models.ManyToManyField('flow.Data')

    #: collection descriptor schema
    descriptor_schema = models.ForeignKey('flow.DescriptorSchema', blank=True, null=True, on_delete=models.PROTECT)

    #: collection descriptor
    descriptor = JSONField(default=dict)

    #: indicate whether `descriptor` doesn't match `descriptor_schema` (is dirty)
    descriptor_dirty = models.BooleanField(default=False)

    #: tags for categorizing objects
    tags = ArrayField(models.CharField(max_length=255), default=list)

    def save(self, *args, **kwargs):
        """Perform descriptor validation and save object."""
        if self.descriptor_schema:
            try:
                validate_schema(self.descriptor, self.descriptor_schema.schema)  # pylint: disable=no-member
                self.descriptor_dirty = False
            except DirtyError:
                self.descriptor_dirty = True
        elif self.descriptor and self.descriptor != {}:
            raise ValueError("`descriptor_schema` must be defined if `descriptor` is given")

        super().save()


class CollectionQuerySet(models.QuerySet):
    """Query set for ``Collection`` objects."""

    @transaction.atomic
    def duplicate(self, contributor=None):
        """Duplicate (make a copy) ``Collection`` objects."""
        return [
            collection.duplicate(contributor=contributor)
            for collection in self
        ]


class Collection(BaseCollection):
    """Postgres model for storing a collection."""

    class Meta(BaseCollection.Meta):
        """Collection Meta options."""

        permissions = (
            ("view_collection", "Can view collection"),
            ("edit_collection", "Can edit collection"),
            ("share_collection", "Can share collection"),
            ("download_collection", "Can download files from collection"),
            ("add_collection", "Can add data objects to collection"),
            ("owner_collection", "Is owner of the collection"),
        )

    #: manager
    objects = CollectionQuerySet.as_manager()

    #: duplication date and time
    duplicated = models.DateTimeField(blank=True, null=True)

    def is_duplicate(self):
        """Return True if collection is a duplicate."""
        return bool(self.duplicated)

    def duplicate(self, contributor=None):
        """Duplicate (make a copy)."""
        duplicate = Collection.objects.get(id=self.id)
        duplicate.pk = None
        duplicate.slug = None
        duplicate.name = 'Copy of {}'.format(self.name)
        duplicate.duplicated = datetime.datetime.now()
        if contributor:
            duplicate.contributor = contributor

        duplicate.save(force_insert=True)  # pylint: disable=protected-access

        assign_contributor_permissions(duplicate)

        # Fields to inherit from original data object.
        duplicate.created = self.created
        duplicate.save()

        # Duplicate collection's entities.
        entities = get_objects_for_user(contributor, 'view_entity', self.entity_set.all())  # pylint: disable=no-member
        duplicated_entities = entities.duplicate(contributor=contributor)
        duplicate.entity_set.add(*duplicated_entities)

        # Add duplicated data objects to collection.
        for duplicated_entity in duplicate.entity_set.all():  # pylint: disable=no-member
            duplicate.data.add(*duplicated_entity.data.all())

        return duplicate



"""Global settings for Flow executor."""
import json
import sys

from .protocol import ExecutorFiles

DESERIALIZED_FILES = {}  # pylint: disable=invalid-name

if 'sphinx' not in sys.modules:
    with open(ExecutorFiles.EXECUTOR_SETTINGS, 'rt') as _settings_file:
        DESERIALIZED_FILES[ExecutorFiles.EXECUTOR_SETTINGS] = json.load(_settings_file)
        for _file_name in DESERIALIZED_FILES[ExecutorFiles.EXECUTOR_SETTINGS][ExecutorFiles.FILE_LIST_KEY]:
            with open(_file_name, 'rt') as _json_file:
                DESERIALIZED_FILES[_file_name] = json.load(_json_file)
else:
    DESERIALIZED_FILES = {getattr(ExecutorFiles, f): {} for f in dir(ExecutorFiles) if f == f.upper()}

EXECUTOR_SETTINGS = DESERIALIZED_FILES[ExecutorFiles.EXECUTOR_SETTINGS]
SETTINGS = DESERIALIZED_FILES[ExecutorFiles.DJANGO_SETTINGS]
DATA = DESERIALIZED_FILES[ExecutorFiles.DATA]
DATA_LOCATION = DESERIALIZED_FILES[ExecutorFiles.DATA_LOCATION]
DATA_META = DESERIALIZED_FILES[ExecutorFiles.DATA_META]
PROCESS = DESERIALIZED_FILES[ExecutorFiles.PROCESS]
PROCESS_META = DESERIALIZED_FILES[ExecutorFiles.PROCESS_META]



"""Safe Resolwe process parser."""
import ast
import collections
import json

import asteval

from . import runtime
from .descriptor import ProcessDescriptor, SchedulingClass
from .fields import get_available_fields


class StaticMetadata:
    """Process metadata item."""

    def get_value(self, node):
        """Convert value from an AST node."""
        raise NotImplementedError


class StaticStringMetadata(StaticMetadata):
    """String metadata item."""

    def get_value(self, node):
        """Convert value from an AST node."""
        if not isinstance(node, ast.Str):
            raise TypeError("must be a string literal")

        return node.s


class StaticEnumMetadata(StaticMetadata):
    """Enum metadata item."""

    def __init__(self, choices):
        """Construct enum metadata item."""
        self.choices = choices

    def get_value(self, node):
        """Convert value from an AST node."""
        if not isinstance(node, ast.Attribute):
            raise TypeError("must be an attribute")

        if node.value.id != self.choices.__name__:
            raise TypeError("must be an attribute of {}".format(self.choices.__name__))

        return getattr(self.choices, node.attr)


class StaticDictMetadata(StaticMetadata):
    """Dict metadata item."""

    def get_value(self, node):
        """Convert value from an AST node."""
        if not isinstance(node, ast.Dict):
            raise TypeError("must be a dictionary")

        evaluator = SafeEvaluator()
        try:
            value = evaluator.run(node)
        except:  # noqa pylint: disable=bare-except
            # TODO: Handle errors.
            raise

        try:
            # Ensure value is a serializable dictionary.
            value = json.loads(json.dumps(value))
            if not isinstance(value, dict):
                raise TypeError
        except (TypeError, ValueError):
            raise TypeError("must be serializable")

        return value


# Possible process metadata.
PROCESS_METADATA = {
    'slug': StaticStringMetadata(),
    'name': StaticStringMetadata(),
    'process_type': StaticStringMetadata(),
    'version': StaticStringMetadata(),
    'category': StaticStringMetadata(),
    'scheduling_class': StaticEnumMetadata(choices=SchedulingClass),
    'requirements': StaticDictMetadata(),
    'data_name': StaticStringMetadata(),
    'entity': StaticDictMetadata(),
}


class SafeEvaluator(asteval.Interpreter):
    """Safe evaluator of Python expressions."""

    def __init__(self, symtable=None):
        """Initialize the safe evaluator."""
        if symtable is None:
            # Ensure symbol table is safely initialized as leaving it set to None makes
            # all kinds of builtin functions available.
            symtable = {}

        super().__init__(
            symtable=symtable,
            use_numpy=False,
            no_print=True,
        )


class ProcessVisitor(ast.NodeVisitor):
    """Visitor for parsing process AST."""

    def __init__(self, source):
        """Construct process AST visitor."""
        self.source = source
        self.processes = []

        super().__init__()

    def visit_field_class(self, item, descriptor=None, fields=None):
        """Visit a class node containing a list of field definitions."""
        discovered_fields = collections.OrderedDict()
        field_groups = {}
        for node in item.body:
            if isinstance(node, ast.ClassDef):
                field_groups[node.name] = self.visit_field_class(node)
                continue

            if not isinstance(node, ast.Assign):
                continue
            if not isinstance(node.value, ast.Call):
                continue
            if not isinstance(node.targets[0], ast.Name):
                continue

            # Build accessible symbols table.
            symtable = {}
            # All field types.
            symtable.update({
                field.__name__: field
                for field in get_available_fields()
            })
            # Field group classes.
            symtable.update(field_groups)

            evaluator = SafeEvaluator(symtable=symtable)

            name = node.targets[0].id
            try:
                field = evaluator.run(node.value)
            except:  # noqa pylint: disable=bare-except
                # TODO: Handle errors.
                raise

            if descriptor is not None:
                field.contribute_to_class(descriptor, fields, name)
            else:
                discovered_fields[name] = field

        if descriptor is None:
            class Fields:
                """Fields wrapper."""

                pass

            for name, field in discovered_fields.items():
                setattr(Fields, name, field)

            return Fields

    def visit_ClassDef(self, node):  # pylint: disable=invalid-name
        """Visit top-level classes."""
        # Resolve everything as root scope contains everything from the process module.
        for base in node.bases:
            # Cover `from resolwe.process import ...`.
            if isinstance(base, ast.Name) and isinstance(base.ctx, ast.Load):
                base = getattr(runtime, base.id, None)
            # Cover `from resolwe import process`.
            elif isinstance(base, ast.Attribute) and isinstance(base.ctx, ast.Load):
                base = getattr(runtime, base.attr, None)
            else:
                continue

            if issubclass(base, runtime.Process):
                break
        else:
            return

        descriptor = ProcessDescriptor(source=self.source)

        # Available embedded classes.
        embedded_class_fields = {
            runtime.PROCESS_INPUTS_NAME: descriptor.inputs,
            runtime.PROCESS_OUTPUTS_NAME: descriptor.outputs,
        }

        # Parse metadata in class body.
        for item in node.body:
            if isinstance(item, ast.Assign):
                # Possible metadata.
                if (len(item.targets) == 1 and isinstance(item.targets[0], ast.Name)
                        and isinstance(item.targets[0].ctx, ast.Store)
                        and item.targets[0].id in PROCESS_METADATA):
                    # Try to get the metadata value.
                    value = PROCESS_METADATA[item.targets[0].id].get_value(item.value)
                    setattr(descriptor.metadata, item.targets[0].id, value)
            elif (isinstance(item, ast.Expr) and isinstance(item.value, ast.Str)
                  and descriptor.metadata.description is None):
                # Possible description string.
                descriptor.metadata.description = item.value.s
            elif isinstance(item, ast.ClassDef) and item.name in embedded_class_fields.keys():
                # Possible input/output declaration.
                self.visit_field_class(item, descriptor, embedded_class_fields[item.name])

        descriptor.validate()
        self.processes.append(descriptor)


class SafeParser:
    """Safe parser for Python processes which doesn't evaluate any code."""

    def __init__(self, source):
        """Construct process parser.

        :param source: Process source code string
        """
        self._source = source

    def parse(self):
        """Parse process.

        :return: A list of discovered process descriptors
        """
        root = ast.parse(self._source)

        visitor = ProcessVisitor(source=self._source)
        visitor.visit(root)
        return visitor.processes



# -*- coding: utf-8 -*-
# Generated by Django 1.11.14 on 2019-03-10 05:36
from __future__ import unicode_literals

from django.db import migrations, models


def mark_purged_locations(apps, schema_editor):
    DataLocation = apps.get_model('flow', 'DataLocation')

    for data_location in DataLocation.objects.all():
        if data_location.data.filter(purged=True).exists():
            data_location.purged = True
            data_location.save()


class Migration(migrations.Migration):

    dependencies = [
        ('flow', '0032_add_collection_duplicate'),
    ]

    operations = [
        migrations.AddField(
            model_name='datalocation',
            name='purged',
            field=models.BooleanField(default=False),
        ),
        migrations.RunPython(mark_purged_locations),
        migrations.RemoveField(
            model_name='data',
            name='purged',
        ),
    ]



#!/usr/bin/env python3
"""Parse tabular file."""
import argparse
import csv
import gzip
import os

import xlrd


def parse_arguments():
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(description="Parse tabular file")
    parser.add_argument('input_file', help="Tabular file.")
    parser.add_argument('output_file', help="Output file.")
    return parser.parse_args()


def main():
    """Run the program."""
    args = parse_arguments()

    ext = os.path.splitext(args.input_file)[-1].lower()
    with gzip.open(args.output_file, mode='wt') as outfile:
        csvwriter = csv.writer(outfile, delimiter=str('\t'), lineterminator='\n')

        try:
            if ext in ('.tab', '.txt', '.tsv'):
                with open(args.input_file) as infile:
                    for line in infile:
                        outline = line.strip().split('\t')
                        csvwriter.writerow(outline)
            elif ext == '.csv':
                with open(args.input_file) as infile:
                    for line in infile:
                        outline = line.strip().split(',')
                        csvwriter.writerow(outline)
            elif ext in ('.xls', '.xlsx'):
                workbook = xlrd.open_workbook(args.input_file)
                worksheet = workbook.sheets()[0]
                for rownum in range(worksheet.nrows):
                    csvwriter.writerow(worksheet.row_values(rownum))
            else:
                print('{"proc.error":"File extension not recognized."}')
        except:  # noqa
            print('{"proc.error":"Corrupt or unrecognized file."}')
            raise


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Generated by Django 1.11.16 on 2019-02-26 04:06
from __future__ import unicode_literals

import os

from django.conf import settings
from django.db import connection, migrations, models, transaction
import django.db.models.deletion


def set_data_location(apps, schema_editor):
    """Create DataLocation for each Data."""
    Data = apps.get_model('flow', 'Data')
    DataLocation = apps.get_model('flow', 'DataLocation')

    for data in Data.objects.all():
        if os.path.isdir(os.path.join(settings.FLOW_EXECUTOR['DATA_DIR'], str(data.id))):
            with transaction.atomic():
                # Manually set DataLocation id to preserve data directory.
                data_location = DataLocation.objects.create(id=data.id, subpath=str(data.id))
                data_location.data.add(data)

    # Increment DataLocation id's sequence
    if DataLocation.objects.exists():
        max_id = DataLocation.objects.order_by('id').last().id
        with connection.cursor() as cursor:
            cursor.execute(
                "ALTER SEQUENCE flow_datalocation_id_seq RESTART WITH {};".format(max_id + 1)
            )


class Migration(migrations.Migration):

    dependencies = [
        ('flow', '0027_data_purged'),
    ]

    operations = [
        migrations.CreateModel(
            name='DataLocation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('subpath', models.CharField(max_length=30, unique=True)),
            ],
        ),
        migrations.AddField(
            model_name='data',
            name='location',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='data', to='flow.DataLocation'),
        ),
        migrations.RunPython(set_data_location, reverse_code=migrations.RunPython.noop),
    ]



""".. Ignore pydocstyle D400.

=====================================
Django Channels Routing Configuration
=====================================

"""

from channels.routing import ChannelNameRouter, ProtocolTypeRouter

from resolwe.flow.managers import state
from resolwe.flow.managers.consumer import ManagerConsumer

channel_routing = ProtocolTypeRouter({  # pylint: disable=invalid-name
    'channel': ChannelNameRouter({
        state.MANAGER_CONTROL_CHANNEL: ManagerConsumer,
    }),
})



""".. Ignore pydocstyle D400.

==================
Permissions Filter
==================

"""
from rest_framework.filters import BaseFilterBackend

from resolwe.permissions.shortcuts import get_objects_for_user


class ResolwePermissionsFilter(BaseFilterBackend):
    """Permissions filter."""

    def filter_queryset(self, request, queryset, view):
        """Filter permissions queryset."""
        user = request.user

        app_label = queryset.model._meta.app_label  # pylint: disable=protected-access
        model_name = queryset.model._meta.model_name  # pylint: disable=protected-access

        kwargs = {}
        if model_name == 'storage':
            model_name = 'data'
            kwargs['perms_filter'] = 'data__pk__in'

        if model_name == 'relation':
            model_name = 'collection'
            kwargs['perms_filter'] = 'collection__pk__in'

        permission = '{}.view_{}'.format(app_label, model_name)

        return get_objects_for_user(user, permission, queryset, **kwargs)



"""Central place for package metadata."""

from pkg_resources import DistributionNotFound, get_distribution

# NOTE: We use __title__ instead of simply __name__ since the latter would
#       interfere with a global variable __name__ denoting object's name.
__title__ = 'resolwe'
__summary__ = 'Open source enterprise dataflow engine in Django'
__url__ = 'https://github.com/genialis/resolwe'

try:
    __version__ = get_distribution(__title__).version
except DistributionNotFound:
    # Package is not (yet) installed.
    pass

__author__ = 'Genialis, Inc.'
__email__ = 'dev-team@genialis.com'

__license__ = 'Apache License (2.0)'
__copyright__ = '2015-2019, ' + __author__

__all__ = (
    '__title__', '__summary__', '__url__', '__version__', '__author__',
    '__email__', '__license__', '__copyright__',
)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.14 on 2019-02-07 02:58
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('flow', '0025_entity_type'),
    ]

    operations = [
        migrations.AddField(
            model_name='data',
            name='scheduled',
            field=models.DateTimeField(blank=True, db_index=True, null=True),
        ),
    ]



"""Query lookup expressions."""
from elasticsearch_dsl.query import Q, Range

# Token separator for parsing lookup expressions.
TOKEN_SEPARATOR = '__'


@staticmethod
def maybe_number(value):
    """Convert number strings to numbers, pass through non-numbers."""
    try:
        return int(value)
    except ValueError:
        try:
            return float(value)
        except ValueError:
            return value


class Lookup:
    """Lookup expression."""

    operator = None
    query_type = None
    value_type = None

    def get_value_query(self, value):
        """Prepare value for use in a query."""
        if self.value_type is None:
            return value

        return self.value_type(value)  # pylint: disable=not-callable

    def apply(self, search, field, value):
        """Apply lookup expression to search query."""
        return search.query(
            self.query_type(  # pylint: disable=not-callable
                **{
                    field: {
                        self.operator: self.get_value_query(value)
                    }
                }
            )
        )


class LessThan(Lookup):
    """Less than lookup."""

    operator = 'lt'
    query_type = Range
    value_type = maybe_number


class LessThanOrEqual(Lookup):
    """Less than or equal lookup."""

    operator = 'lte'
    query_type = Range
    value_type = maybe_number


class GreaterThan(Lookup):
    """Greater than lookup."""

    operator = 'gt'
    query_type = Range
    value_type = maybe_number


class GreaterThanOrEqual(Lookup):
    """Greater than or equal lookup."""

    operator = 'gte'
    query_type = Range
    value_type = maybe_number


class In(Lookup):
    """In lookup."""

    operator = 'in'

    def apply(self, search, field, value):
        """Apply lookup expression to search query."""
        if not isinstance(value, list):
            value = [x for x in value.strip().split(',') if x]

        filters = [Q('match', **{field: item}) for item in value]
        return search.query('bool', should=filters)


class Exact(Lookup):
    """Exact lookup."""

    operator = 'exact'

    def apply(self, search, field, value):
        """Apply lookup expression to search query."""
        # We assume that the field in question has a "raw" counterpart.
        return search.query('match', **{'{}.raw'.format(field): value})


class QueryBuilder:
    """Query builder."""

    def __init__(self, fields, fields_map, custom_filter_object):
        """Construct query builder.

        :param fields: List of supported fields
        :param fields_map: Field alias mapping
        :param custom_filter_object: Object for calling custom overrides
        """
        self.fields = fields
        self.fields_map = fields_map
        self.custom_filter_object = custom_filter_object
        self._lookups = {}

        # Register default lookups.
        self.register_lookup(LessThan)
        self.register_lookup(LessThanOrEqual)
        self.register_lookup(GreaterThan)
        self.register_lookup(GreaterThanOrEqual)
        self.register_lookup(In)
        self.register_lookup(Exact)

    def register_lookup(self, lookup):
        """Register lookup."""
        if lookup.operator in self._lookups:
            raise KeyError("Lookup for operator '{}' is already registered".format(lookup.operator))

        self._lookups[lookup.operator] = lookup()

    def get_lookup(self, operator):
        """Look up a lookup.

        :param operator: Name of the lookup operator
        """
        try:
            return self._lookups[operator]
        except KeyError:
            raise NotImplementedError("Lookup operator '{}' is not supported".format(operator))

    def build(self, search, raw_query):
        """Build query.

        :param search: Search query instance
        :param raw_query: Raw query arguments dictionary
        """
        unmatched_items = {}
        for expression, value in raw_query.items():
            # Parse query expression into tokens.
            tokens = expression.split(TOKEN_SEPARATOR)

            field = tokens[0]
            tail = tokens[1:]

            if field not in self.fields:
                unmatched_items[expression] = value
                continue

            # Map field alias to final field.
            field = self.fields_map.get(field, field)

            # Parse lookup expression. Currently only no token or a single token is allowed.
            if tail:
                if len(tail) > 1:
                    raise NotImplementedError("Nested lookup expressions are not supported")

                lookup = self.get_lookup(tail[0])
                search = lookup.apply(search, field, value)
            else:
                # Default lookup.
                custom_filter = getattr(self.custom_filter_object, 'custom_filter_{}'.format(field), None)
                if custom_filter is not None:
                    search = custom_filter(value, search)
                elif isinstance(value, list):
                    # Default is 'should' between matches. If you need anything else,
                    # a custom filter for this field should be implemented.
                    filters = [Q('match', **{field: item}) for item in value]
                    search = search.query('bool', should=filters)
                else:
                    search = search.query('match', **{field: {'query': value, 'operator': 'and'}})

        return (search, unmatched_items)



# pylint: disable=missing-docstring
import os

from resolwe.flow.models import Entity, Process
from resolwe.test import ProcessTestCase, tag_process, with_docker_executor

PROCESSES_DIR = os.path.join(os.path.dirname(__file__), 'processes')
DESCRIPTORS_DIR = os.path.join(os.path.dirname(__file__), 'descriptors')
FILES_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'files')


class PythonProcessTest(ProcessTestCase):
    def setUp(self):
        super().setUp()
        self._register_schemas(processes_paths=[PROCESSES_DIR], descriptors_paths=[DESCRIPTORS_DIR])
        self.files_path = FILES_PATH

    @with_docker_executor
    def test_registration(self):
        process = Process.objects.get(slug='test-python-process')

        self.assertEqual(process.slug, 'test-python-process')
        self.assertEqual(process.name, 'Test Python Process')
        self.assertEqual(process.version, '0.1.2')
        self.assertEqual(process.type, 'data:python:')
        self.assertEqual(process.category, 'analyses:')
        self.assertEqual(process.persistence, Process.PERSISTENCE_RAW)
        self.assertEqual(process.description, 'This is a process description.')
        self.assertEqual(process.data_name, 'Foo: {{input_data | name}}')
        self.assertEqual(process.entity_type, 'sample')
        self.assertEqual(process.entity_descriptor_schema, 'sample')
        self.assertEqual(process.entity_input, 'input_data')
        self.assertEqual(process.requirements, {
            'expression-engine': 'jinja',
            'executor': {
                'docker': {
                    'image': 'resolwe/base:ubuntu-18.04',
                }
            }
        })
        self.assertEqual(process.scheduling_class, Process.SCHEDULING_CLASS_BATCH)

        # Make sure that process with inheritance from `module.Class` is also registered.
        process = Process.objects.get(slug='test-python-process-2')

    @with_docker_executor
    @tag_process('test-python-process')
    def test_python_process(self):
        with self.preparation_stage():
            input_data = self.run_process('test-save-number', {'number': 19})
            input_data.name = "bar"
            input_data.save()

            storage = input_data.storages.create(
                name="storage",
                contributor=self.user,
                json={'value': 42}
            )

        data = self.run_process('test-python-process', {
            'my_field': "bar",
            'my_list': ["one", "two", "three"],
            'bar': input_data.pk,
            'url': {'url': "https://www.genialis.com"},
            'input_data': input_data.pk,
            'integer': 42,
            'my_float': 0.42,
            'my_json': storage.pk,
            'my_group': {
                'bar': 'my string',
                'foo': 21,
            }
        })

        self.assertFields(data, 'my_output', 'OK')
        self.assertFile(data, 'file_output', 'testfile.txt')

        # Non-deterministic output.
        del data.output['dir_output']['size']
        del data.output['dir_output']['total_size']

        self.assertEqual(data.output['dir_output'], {'dir': 'test/'})

        self.assertEqual(data.name, 'Foo: bar')

        entity = Entity.objects.get(data=data)
        self.assertEqual(entity.data.first(), data)

    @with_docker_executor
    @tag_process('test-python-process-json')
    def test_python_process_json(self):
        """Test that data object with json output can be given as input."""
        with self.preparation_stage():
            input_data = self.run_process('test-output-json')

        self.run_process('test-python-process-json', {
            'data': input_data.pk,
        })



"""Resolwe process serializer."""
from resolwe.flow.models import Process

from .base import ResolweBaseSerializer


class ProcessSerializer(ResolweBaseSerializer):
    """Serializer for Process objects."""

    class Meta:
        """ProcessSerializer Meta options."""

        model = Process
        read_only_fields = (
            'created',
            'id',
            'modified',
            'is_active',
        )
        update_protected_fields = (
            'category',
            'contributor',
            'data_name',
            'description',
            'entity_descriptor_schema',
            'entity_input',
            'entity_type',
            'input_schema',
            'name',
            'output_schema',
            'persistence',
            'requirements',
            'run',
            'scheduling_class',
            'slug',
            'type',
            'version',
        )
        fields = read_only_fields + update_protected_fields



from resolwe import process
from resolwe.process import *


class PythonProcess(Process):
    """This is a process description."""
    slug = 'test-python-process'
    name = "Test Python Process"
    version = '0.1.2'
    process_type = 'data:python'
    category = 'analyses'
    scheduling_class = SchedulingClass.BATCH
    data_name = "Foo: {{input_data | name}}"
    entity = {
        'type': 'sample',
        'descriptor_schema': 'sample',
        'input': 'input_data',
    }
    requirements = {
        'expression-engine': 'jinja',
        'executor': {
            'docker': {
                'image': 'resolwe/base:ubuntu-18.04',
            }
        }
    }

    class Input:
        """Input fields."""
        my_field = StringField(label="My field")
        my_list = ListField(StringField(), label="My list")
        input_data = DataField('test:save', label="My input data")
        bar = DataField(data_type='test:save', label="My bar")
        url = UrlField(UrlField.DOWNLOAD, label="My URL")
        integer = IntegerField(label="My integer")
        my_float = FloatField(label="My float")
        my_json = JsonField(label="Blah blah")
        my_optional = StringField(label="Optional", required=False, default='default value')

        class MyGroup:
            bar = StringField(label="Bar")
            foo = IntegerField(label="Foo")

        my_group = GroupField(MyGroup, label="My group")

    class Output:
        my_output = StringField(label="My output")
        file_output = FileField(label="My output")
        dir_output = DirField(label="My output")

    def run(self, inputs, outputs):
        print('All inputs are:', inputs)
        print('Input data:', inputs.input_data)
        print('Input data ID:', inputs.input_data.id)
        print('Input data number output:', inputs.input_data.number)
        print('Input data type:', inputs.input_data.type)
        print('Input data descriptor:', inputs.input_data.descriptor)
        print('Group bar:', inputs.my_group.bar)
        print('Group foo:', inputs.my_group.foo)

        bar = Cmd['ls']['-l', '-a', '/'] | Cmd['grep']['python']
        print('hello world:\n', bar())

        cmd = Cmd['mkdir']['test']()
        cmd = (Cmd['echo']['"Some content"'] > 'test/testfile.txt')()
        outputs.file_output = 'test/testfile.txt'
        outputs.dir_output = 'test/'

        outputs.my_output = 'OK'


class PythonProcess2(process.Process):
    """Inherit from 'module.Class'."""
    slug = 'test-python-process-2'
    name = "Test Python Process 2"
    version = '0.0.1'
    process_type = 'data:python'



# -*- coding: utf-8 -*-
# Generated by Django 1.11.6 on 2017-11-20 03:49
from __future__ import unicode_literals

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('flow', '0012_recreate_empty_parents'),
    ]

    operations = [
        migrations.CreateModel(
            name='DataMigrationHistory',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('migration', models.CharField(db_index=True, max_length=255)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('metadata', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('data', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='migration_history', to='flow.Data')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='ProcessMigrationHistory',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('migration', models.CharField(db_index=True, max_length=255)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('metadata', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('process', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='migration_history', to='flow.Process')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



"""Process viewset."""
from rest_framework import exceptions, mixins, viewsets

from resolwe.flow.filters import ProcessFilter
from resolwe.flow.models import Process
from resolwe.flow.serializers import ProcessSerializer
from resolwe.permissions.loader import get_permissions_class
from resolwe.permissions.mixins import ResolwePermissionsMixin

from .mixins import ResolweCheckSlugMixin, ResolweCreateModelMixin


class ProcessViewSet(ResolweCreateModelMixin,
                     mixins.RetrieveModelMixin,
                     mixins.ListModelMixin,
                     ResolwePermissionsMixin,
                     ResolweCheckSlugMixin,
                     viewsets.GenericViewSet):
    """API view for :class:`Process` objects."""

    queryset = Process.objects.all().prefetch_related('contributor')
    serializer_class = ProcessSerializer
    permission_classes = (get_permissions_class(),)
    filter_class = ProcessFilter
    ordering_fields = ('id', 'created', 'modified', 'name', 'version')
    ordering = ('id',)

    def create(self, request, *args, **kwargs):
        """Only superusers can create new processes."""
        if not request.user.is_superuser:
            raise exceptions.NotFound

        return super().create(request, *args, **kwargs)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.6 on 2017-10-20 08:00
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('flow', '0006_add_total_size'),
    ]

    operations = [
        migrations.AddField(
            model_name='data',
            name='size',
            field=models.BigIntegerField(blank=True, null=True),
        ),
    ]



# pylint: disable=missing-docstring
import elasticsearch_dsl as dsl

from resolwe.elastic.builder import ManyToManyDependency
from resolwe.elastic.fields import Name, ProcessType
from resolwe.elastic.indices import BaseDocument, BaseIndex

from .models import TestDependency, TestModel, TestModelWithDependency, TestSelfDependency


class TestSearchDocument(BaseDocument):
    # pylint: disable=no-member
    id = dsl.Integer()  # pylint: disable=invalid-name
    name = dsl.Text(fielddata=True)
    num = dsl.Integer()
    date = dsl.Date()
    json = dsl.Object()

    field_name = Name()
    field_process_type = ProcessType()
    none_test = dsl.Integer()

    class Meta:
        index = 'test_search'


class TestSearchIndex(BaseIndex):
    mapping = {
        'num': 'number',
        'field_name': 'name',
        'none_test': 'this.does.not.exist',
    }

    queryset = TestModel.objects.all()
    object_type = TestModel
    document_class = TestSearchDocument

    def get_json_value(self, obj):
        return {'key': 'value'}


class TestAnalyzerSearchDocument(BaseDocument):
    # pylint: disable=no-member
    name = dsl.Text(analyzer=dsl.analyzer(
        'test_analyzer',
        tokenizer='keyword',
        filter=[
            'lowercase',
        ],
        fielddata=True,
    ))

    class Meta:
        index = 'test_analyzer_search'


class TestAnalyzerSearchIndex(BaseIndex):
    queryset = TestModel.objects.all()
    object_type = TestModel
    document_class = TestAnalyzerSearchDocument


class TestModelWithDependencyDocument(BaseDocument):
    # pylint: disable=no-member
    name = dsl.Text(fielddata=True)

    class Meta:
        index = 'test_model_with_dependency_search'


class TestModelWithDependencySearchIndex(BaseIndex):
    queryset = TestModelWithDependency.objects.all().prefetch_related('dependencies')
    object_type = TestModelWithDependency
    document_class = TestModelWithDependencyDocument

    def get_dependencies(self):
        return [TestModelWithDependency.dependencies]

    def get_name_value(self, obj):
        names = [dep.name for dep in obj.dependencies.all()]
        return '{}: {}'.format(obj.name, ', '.join(names))


class TestModelWithFilterDependencyDocument(BaseDocument):
    # pylint: disable=no-member
    name = dsl.Text(fielddata=True)

    class Meta:
        index = 'test_model_with_filter_dependency_search'


class FilterHelloDependency(ManyToManyDependency):
    def filter(self, obj, update_fields=None):
        return obj.name == 'hello'


class TestModelWithFilterDependencySearchIndex(BaseIndex):
    queryset = TestModelWithDependency.objects.all().prefetch_related('dependencies')
    object_type = TestModelWithDependency
    document_class = TestModelWithFilterDependencyDocument

    def get_dependencies(self):
        return [FilterHelloDependency(TestModelWithDependency.dependencies)]

    def get_name_value(self, obj):
        names = [dep.name for dep in obj.dependencies.all()]
        return '{}: {}'.format(obj.name, ', '.join(names))


class TestModelWithReverseDependencyDocument(BaseDocument):
    # pylint: disable=no-member
    name = dsl.String()

    class Meta:
        index = 'test_model_with_reverse_dependency_search'


class TestModelWithReverseDependencySearchIndex(BaseIndex):
    queryset = TestDependency.objects.all()
    object_type = TestDependency
    document_class = TestModelWithReverseDependencyDocument

    def get_dependencies(self):
        return [TestDependency.testmodelwithdependency_set]  # pylint: disable=no-member

    def get_name_value(self, obj):
        names = [dep.name for dep in obj.testmodelwithdependency_set.all()]
        return '{}: {}'.format(obj.name, ', '.join(names))


class TestModelWithSelfDependencyDocument(BaseDocument):
    # pylint: disable=no-member
    name = dsl.String()

    class Meta:
        index = 'test_model_with_self_dependency_search'


class TestModelWithSelfDependencySearchIndex(BaseIndex):
    queryset = TestSelfDependency.objects.all()
    object_type = TestSelfDependency
    document_class = TestModelWithSelfDependencyDocument

    def get_dependencies(self):
        return [TestSelfDependency.dependencies]

    def get_name_value(self, obj):
        names = [dep.name for dep in obj.dependencies.all().order_by('pk')]
        return '{}: {}'.format(obj.name, ', '.join(names))



"""Process input or output fields."""
import collections
import json

import resolwe_runtime_utils


class ValidationError(Exception):
    """Field value validation error."""

    pass


class Field:
    """Process input or output field."""

    field_type = None

    def __init__(self, label=None, required=True, description=None, default=None, choices=None,
                 hidden=False):
        """Construct a field descriptor."""
        self.name = None
        self.process = None
        self.label = label
        self.required = required
        self.description = description
        self.default = default
        self.choices = choices
        self.hidden = hidden

    def get_field_type(self):
        """Return this field's type."""
        return self.field_type

    def contribute_to_class(self, process, fields, name):
        """Register this field with a specific process.

        :param process: Process descriptor instance
        :param fields: Fields registry to use
        :param name: Field name
        """
        self.name = name
        self.process = process
        fields[name] = self

    def to_python(self, value):
        """Convert value if needed."""
        return value

    def to_schema(self):
        """Return field schema for this field."""
        if not self.name or not self.process:
            raise ValueError("field is not registered with process")

        schema = {
            'name': self.name,
            'type': self.get_field_type(),
        }
        if self.required is not None:
            schema['required'] = self.required
        if self.label is not None:
            schema['label'] = self.label
        if self.description is not None:
            schema['description'] = self.description
        if self.default is not None:
            schema['default'] = self.default
        if self.hidden is not None:
            schema['hidden'] = self.hidden
        if self.choices is not None:
            for choice, label in self.choices:
                schema.setdefault('choices', []).append({
                    'label': label,
                    'value': choice,
                })

        return schema

    def to_output(self, value):
        """Convert value to process output format."""
        return json.loads(resolwe_runtime_utils.save(self.name, value))

    def validate(self, value):
        """Validate field value."""
        if self.required and value is None:
            raise ValidationError("field is required")

        if value is not None and self.choices is not None:
            choices = [choice for choice, _ in self.choices]
            if value not in choices:
                raise ValidationError("field must be one of: {}".format(
                    ", ".join(choices),
                ))

    def clean(self, value):
        """Run validators and return the clean value."""
        if value is None:
            value = self.default

        try:
            value = self.to_python(value)
            self.validate(value)
        except ValidationError as error:
            raise ValidationError("invalid value for {}: {}".format(
                self.name,
                error.args[0]
            ))
        return value

    def __repr__(self):
        """Return string representation."""
        return '<{klass} name={name} type={type} label="{label}">'.format(
            klass=self.__class__.__name__,
            name=self.name,
            type=self.get_field_type(),
            label=self.label,
        )


class StringField(Field):
    """String field."""

    field_type = 'basic:string'

    def validate(self, value):
        """Validate field value."""
        if value is not None and not isinstance(value, str):
            raise ValidationError("field must be a string")

        super().validate(value)


class TextField(StringField):
    """Text field."""

    field_type = 'basic:text'


class BooleanField(Field):
    """Boolean field."""

    field_type = 'basic:boolean'

    def validate(self, value):
        """Validate field value."""
        if value is not None and not isinstance(value, bool):
            raise ValidationError("field must be a boolean")

        super().validate(value)


class IntegerField(Field):
    """Integer field."""

    field_type = 'basic:integer'

    def to_python(self, value):
        """Convert value if needed."""
        if value is not None:
            try:
                return int(value)
            except (TypeError, ValueError):
                raise ValidationError("field must be an integer")


class FloatField(Field):
    """Float field."""

    # TODO: Fix the underlying field into basic:float once that is renamed.
    field_type = 'basic:decimal'

    def to_python(self, value):
        """Convert value if needed."""
        if value is not None:
            try:
                return float(value)
            except (TypeError, ValueError):
                raise ValidationError("field must be a float")


class DateField(Field):
    """Date field."""

    field_type = 'basic:date'


class DateTimeField(Field):
    """Date time field."""

    field_type = 'basic:datetime'


class UrlField(Field):
    """URL field."""

    # Url types.
    DOWNLOAD = 'download'
    VIEW = 'view'
    LINK = 'link'

    URL_TYPES = (DOWNLOAD, VIEW, LINK)

    def __init__(self, url_type, *args, **kwargs):
        """Construct an URL field descriptor.

        :param url_type: Type of URL
        """
        if url_type not in self.URL_TYPES:
            raise ValueError("url_type must be one of: {}".format(', '.join(self.URL_TYPES)))

        self.url_type = url_type
        super().__init__(*args, **kwargs)

    def to_python(self, value):
        """Convert value if needed."""
        if isinstance(value, str):
            return value
        elif isinstance(value, dict):
            try:
                value = value['url']
            except KeyError:
                raise ValidationError("dictionary must contain an 'url' element")

            if not isinstance(value, str):
                raise ValidationError("field's url element must be a string")

            return value
        elif not isinstance(value, None):
            raise ValidationError("field must be a string or a dict")

    def get_field_type(self):
        """Return this field's type."""
        return 'basic:url:{}'.format(self.url_type)


class SecretField(Field):
    """Secret field."""

    field_type = 'basic:secret'


class FileDescriptor:
    """Descriptor for accessing files."""

    def __init__(self, path, size=None, total_size=None, is_remote=False, file_temp=None, refs=None):
        """Construct a file descriptor."""
        self.path = path
        self.size = size
        self.total_size = total_size
        self.is_remote = is_remote
        self.file_temp = file_temp
        if refs is None:
            refs = []
        self.refs = refs

    def __repr__(self):
        """Return string representation."""
        return '<FileDescriptor path={}>'.format(self.path)


class FileField(Field):
    """File field."""

    field_type = 'basic:file'

    def to_python(self, value):
        """Convert value if needed."""
        if isinstance(value, FileDescriptor):
            return value
        elif isinstance(value, str):
            return FileDescriptor(value)
        elif isinstance(value, dict):
            try:
                path = value['file']
            except KeyError:
                raise ValidationError("dictionary must contain a 'file' element")

            if not isinstance(path, str):
                raise ValidationError("field's file element must be a string")

            size = value.get('size', None)
            if size is not None and not isinstance(size, int):
                raise ValidationError("field's size element must be an integer")

            total_size = value.get('total_size', None)
            if total_size is not None and not isinstance(total_size, int):
                raise ValidationError("field's total_size element must be an integer")

            is_remote = value.get('is_remote', None)
            if is_remote is not None and not isinstance(is_remote, bool):
                raise ValidationError("field's is_remote element must be a boolean")

            file_temp = value.get('file_temp', None)
            if file_temp is not None and not isinstance(file_temp, str):
                raise ValidationError("field's file_temp element must be a string")

            refs = value.get('refs', None)
            if refs is not None and not isinstance(refs, list):
                # TODO: Validate that all refs are strings.
                raise ValidationError("field's refs element must be a list of strings")

            return FileDescriptor(
                path,
                size=size,
                total_size=total_size,
                is_remote=is_remote,
                file_temp=file_temp,
                refs=refs,
            )
        elif not isinstance(value, None):
            raise ValidationError("field must be a FileDescriptor, string or a dict")

    def to_output(self, value):
        """Convert value to process output format."""
        return json.loads(resolwe_runtime_utils.save_file(self.name, value.path, *value.refs))


class FileHtmlField(FileField):
    """HTML file field."""

    field_type = 'basic:file:html'


class DirDescriptor:
    """Descriptor for accessing directories."""

    def __init__(self, path, size=None, total_size=None, refs=None):
        """Construct a file descriptor."""
        self.path = path
        self.size = size
        self.total_size = total_size
        if refs is None:
            refs = []
        self.refs = refs

    def __repr__(self):
        """Return string representation."""
        return '<DirDescriptor path={}>'.format(self.path)


class DirField(Field):
    """Directory field."""

    field_type = 'basic:dir'

    def to_python(self, value):
        """Convert value if needed."""
        if isinstance(value, DirDescriptor):
            return value
        elif isinstance(value, str):
            return DirDescriptor(value)
        elif isinstance(value, dict):
            try:
                path = value['dir']
            except KeyError:
                raise ValidationError("dictionary must contain a 'dir' element")

            if not isinstance(path, str):
                raise ValidationError("field's dir element must be a string")

            size = value.get('size', None)
            if size is not None and not isinstance(size, int):
                raise ValidationError("field's size element must be an integer")

            total_size = value.get('total_size', None)
            if total_size is not None and not isinstance(total_size, int):
                raise ValidationError("field's total_size element must be an integer")

            refs = value.get('refs', None)
            if refs is not None and not isinstance(refs, list):
                # TODO: Validate that all refs are strings.
                raise ValidationError("field's refs element must be a list of strings")

            return DirDescriptor(
                path,
                size=size,
                total_size=total_size,
                refs=refs,
            )
        elif not isinstance(value, None):
            raise ValidationError("field must be a DirDescriptor, string or a dict")

    def to_output(self, value):
        """Convert value to process output format."""
        return json.loads(resolwe_runtime_utils.save_dir(self.name, value.path, *value.refs))


class JsonField(Field):
    """JSON field."""

    field_type = 'basic:json'

    def to_python(self, value):
        """Convert value if needed."""
        if value is not None:
            return json.loads(json.dumps(value))


class ListField(Field):
    """Generic list field."""

    def __init__(self, inner, *args, **kwargs):
        """Construct a list field."""
        if not isinstance(inner, Field):
            raise TypeError("inner field must be an instance of Field")

        self.inner = inner
        super().__init__(*args, **kwargs)

    def contribute_to_class(self, process, fields, name):
        """Register this field with a specific process.

        :param process: Process descriptor instance
        :param fields: Fields registry to use
        :param name: Field name
        """
        super().contribute_to_class(process, fields, name)

        self.inner.name = name
        self.inner.process = process

    def to_output(self, value):
        """Convert value to process output format."""
        return {self.name: [value[self.name] for value in value]}

    def get_field_type(self):
        """Return this field's type."""
        return 'list:{}'.format(self.inner.get_field_type())

    def validate(self, value):
        """Validate field value."""
        if value is not None:
            if not isinstance(value, list):
                raise ValidationError("field must be a list")

            for index, element in enumerate(value):
                try:
                    self.inner.validate(element)
                except ValidationError as error:
                    raise ValidationError("invalid element {}: {}".format(
                        index,
                        error.args[0],
                    ))

        super().validate(value)


class DataDescriptor:
    """Descriptor for accessing data objects."""

    def __init__(self, data_id, field, cache=None):
        """Construct a data descriptor.

        :param data_id: Data object primary key
        :param field: Field this descriptor is for
        :param cache: Optional cached object to use
        """
        super().__setattr__('_data_id', data_id)
        super().__setattr__('_cache', cache)

    def _populate_cache(self):
        """Fetch data object from the backend if needed."""
        if self._data_id is None:
            return

        if self._cache is None:
            # TODO: Implement fetching via the protocol once available.
            raise NotImplementedError

    def _get(self, key):
        """Return given key from cache."""
        self._populate_cache()
        try:
            return self._cache[key]
        except KeyError:
            raise AttributeError(key)

    @property
    def id(self):  # pylint: disable=invalid-name
        """Primary key of this data object."""
        return self._data_id

    @property
    def type(self):
        """Type of this data object."""
        return self._get('__type')

    @property
    def descriptor(self):
        """Descriptor of this data object."""
        return self._get('__descriptor')

    def __getattr__(self, key):
        """Get attribute."""
        return self._get(key)

    def __setattr__(self, key, value):
        """Set attribute."""
        raise AttributeError("inputs are read-only")

    def __repr__(self):
        """Return string representation."""
        return '<DataDescriptor id={}>'.format(self._data_id)


class DataField(Field):
    """Data object field."""

    def __init__(self, data_type, *args, **kwargs):
        """Construct a data field."""
        # TODO: Validate data type format.
        self.data_type = data_type
        super().__init__(*args, **kwargs)

    def get_field_type(self):
        """Return this field's type."""
        return 'data:{}'.format(self.data_type)

    def to_python(self, value):
        """Convert value if needed."""
        cache = None
        if value is not None:
            if isinstance(value, DataDescriptor):
                return value
            elif isinstance(value, dict):
                # Allow pre-hydrated data objects.
                cache = value
                try:
                    value = cache['__id']
                except KeyError:
                    raise ValidationError("dictionary must contain an '__id' element")
            elif not isinstance(value, int):
                raise ValidationError("field must be a DataDescriptor, data object primary key or dict")

        return DataDescriptor(value, self, cache=cache)


class GroupDescriptor:
    """Group descriptor."""

    def __init__(self, value):
        """Construct a group descriptor."""
        self._value = value

    def __getattr__(self, name):
        """Get attribute."""
        try:
            return self._value[name]
        except KeyError:
            raise AttributeError(name)


class GroupField(Field):
    """Group field."""

    field_type = 'basic:group'

    def __init__(self, field_group, label=None, description=None, disabled=False, collapsed=False,
                 hidden=False):
        """Construct a group field."""
        super().__init__(
            label=label,
            required=None,
            description=description,
            hidden=hidden,
        )

        self.disabled = disabled
        self.collapsed = collapsed
        self.field_group = field_group
        self.fields = collections.OrderedDict()

    def contribute_to_class(self, process, fields, name):
        """Register this field with a specific process.

        :param process: Process descriptor instance
        :param fields: Fields registry to use
        :param name: Field name
        """
        for field_name in dir(self.field_group):
            if field_name.startswith('_'):
                continue

            field = getattr(self.field_group, field_name)
            field.contribute_to_class(process, self.fields, field_name)

        super().contribute_to_class(process, fields, name)

    def to_python(self, value):
        """Convert value if needed."""
        if isinstance(value, GroupDescriptor):
            value = value._value  # pylint: disable=protected-access

        result = {}
        for name, field in self.fields.items():
            result[name] = field.to_python(value.get(name, None))

        return GroupDescriptor(result)

    def to_schema(self):
        """Return field schema for this field."""
        schema = super().to_schema()
        if self.disabled is not None:
            schema['disabled'] = self.disabled
        if self.collapsed is not None:
            schema['collapsed'] = self.collapsed

        group = []
        for field in self.fields.values():
            group.append(field.to_schema())
        schema['group'] = group

        return schema


# List of available fields.
ALL_FIELDS = [
    StringField,
    TextField,
    BooleanField,
    IntegerField,
    FloatField,
    DateField,
    DateTimeField,
    UrlField,
    SecretField,
    FileField,
    FileHtmlField,
    DirField,
    JsonField,
    ListField,
    DataField,
    GroupField,
]


def get_available_fields():
    """Return a list of available field classes."""
    return ALL_FIELDS



""".. Ignore pydocstyle D400.

===========
Preparation
===========

.. autoclass:: resolwe.flow.executors.docker.prepare.FlowExecutorPreparer
    :members:

"""

import os

from django.conf import settings
from django.core.management import call_command

from ..prepare import BaseFlowExecutorPreparer
from . import constants


class FlowExecutorPreparer(BaseFlowExecutorPreparer):
    """Specialized manager assist for the docker executor."""

    def post_register_hook(self, verbosity=1):
        """Pull Docker images needed by processes after registering."""
        if not getattr(settings, 'FLOW_DOCKER_DONT_PULL', False):
            call_command('list_docker_images', pull=True, verbosity=verbosity)

    def resolve_data_path(self, data=None, filename=None):
        """Resolve data path for use with the executor.

        :param data: Data object instance
        :param filename: Filename to resolve
        :return: Resolved filename, which can be used to access the
            given data file in programs executed using this executor
        """
        if data is None:
            return constants.DATA_ALL_VOLUME

        # Prefix MUST be set because ``get_path`` uses Django's settings,
        # if prefix is not set, to get path prefix. But the executor
        # shouldn't use Django's settings directly, so prefix is set
        # via a constant.
        return data.location.get_path(prefix=constants.DATA_ALL_VOLUME, filename=filename)

    def resolve_upload_path(self, filename=None):
        """Resolve upload path for use with the executor.

        :param filename: Filename to resolve
        :return: Resolved filename, which can be used to access the
            given uploaded file in programs executed using this
            executor
        """
        if filename is None:
            return constants.UPLOAD_VOLUME

        return os.path.join(constants.UPLOAD_VOLUME, filename)

    def get_environment_variables(self):
        """Return dict of environment variables that will be added to executor."""
        return {
            'TMPDIR': os.path.join(constants.DATA_VOLUME, constants.TMPDIR)
        }



"""Process descriptor."""
import collections
import re

# Processor type validation regexp.
PROCESSOR_TYPE_RE = re.compile(r'^data:[a-z0-9:]+$')


class SchedulingClass:
    """Process scheduling class."""

    INTERACTIVE = 'interactive'
    BATCH = 'batch'


class ValidationError(Exception):
    """Process descriptor validation error."""

    pass


class ProcessDescriptor:
    """Process descriptor.

    The process descriptor is a result of parsing a process definition. It contains
    everything that is required for registering the process into the database.
    """

    class Metadata:
        """General process metadata."""

        slug = None
        name = None
        process_type = None
        description = None
        version = None
        category = None
        scheduling_class = None
        requirements = None
        data_name = None
        entity = None

        def __repr__(self):
            """Return string representation."""
            return repr(self.__dict__)

    def __init__(self, source=None):
        """Construct a new process descriptor."""
        self.metadata = ProcessDescriptor.Metadata()
        self.inputs = collections.OrderedDict()
        self.outputs = collections.OrderedDict()
        self.source = source

    def validate(self):
        """Validate process descriptor."""
        required_fields = ('slug', 'name', 'process_type', 'version')
        for field in required_fields:
            if getattr(self.metadata, field, None) is None:
                raise ValidationError("process '{}' is missing required meta attribute: {}".format(
                    self.metadata.slug or '<unknown>', field))

        if not PROCESSOR_TYPE_RE.match(self.metadata.process_type):
            raise ValidationError("process '{}' has invalid type: {}".format(
                self.metadata.slug, self.metadata.process_type))

    def to_schema(self):
        """Return process schema for this process."""
        process_type = self.metadata.process_type
        if not process_type.endswith(':'):
            process_type = '{}:'.format(process_type)

        schema = {
            'slug': self.metadata.slug,
            'name': self.metadata.name,
            'type': process_type,
            'version': self.metadata.version,
            'data_name': '',
            'requirements': {
                'executor': {
                    'docker': {
                        'image': 'resolwe/base:ubuntu-18.10',
                    },
                },
            },
        }

        if self.metadata.description is not None:
            schema['description'] = self.metadata.description
        if self.metadata.category is not None:
            schema['category'] = self.metadata.category
        if self.metadata.scheduling_class is not None:
            schema['scheduling_class'] = self.metadata.scheduling_class
        if self.metadata.requirements is not None:
            schema['requirements'] = self.metadata.requirements
        if self.metadata.data_name is not None:
            schema['data_name'] = self.metadata.data_name
        if self.metadata.entity is not None:
            schema['entity'] = self.metadata.entity

        if self.inputs:
            schema['input'] = []
            for field in self.inputs.values():
                schema['input'].append(field.to_schema())

        if self.outputs:
            schema['output'] = []
            for field in self.outputs.values():
                schema['output'].append(field.to_schema())

        schema['run'] = {
            'language': 'python',
            'program': self.source or '',
        }

        return schema

    def __repr__(self):
        """Return string representation."""
        return '<ProcessDescriptor metadata={} inputs={} outputs={}>'.format(
            repr(self.metadata),
            repr(dict(self.inputs)),
            repr(dict(self.outputs)),
        )



"""Main standalone execution stub, used when the executor is run.

It should be run as a module with one argument: the relative module name
of the concrete executor class to use. The current working directory
should be where the ``executors`` module directory is, so that it can be
imported with python's ``-m <module>`` interpreter option.

Usage format:

.. code-block:: none

    /path/to/python -m executors .executor_type

Concrete example, run from the directory where ``./executors/`` is:

.. code-block:: none

    /venv/bin/python -m executors .docker

using the python from the ``venv`` virtualenv.

.. note::

    The startup code adds the concrete class name as needed, so that in
    the example above, what's actually instantiated is
    ``.docker.run.FlowExecutor``.
"""

import argparse
import asyncio
from importlib import import_module

from . import manager_commands
from .global_settings import DATA
from .logger import configure_logging
from .protocol import ExecutorFiles


async def run_executor():
    """Start the actual execution; instantiate the executor and run."""
    parser = argparse.ArgumentParser(description="Run the specified executor.")
    parser.add_argument('module', help="The module from which to instantiate the concrete executor.")
    args = parser.parse_args()

    module_name = '{}.run'.format(args.module)
    class_name = 'FlowExecutor'

    module = import_module(module_name, __package__)
    executor = getattr(module, class_name)()
    with open(ExecutorFiles.PROCESS_SCRIPT, 'rt') as script_file:
        await executor.run(DATA['id'], script_file.read())


if __name__ == '__main__':
    logging_future_list = []  # pylint: disable=invalid-name
    configure_logging(logging_future_list)

    async def _sequential():
        """Run some things sequentially but asynchronously."""
        await manager_commands.init()
        await run_executor()
    loop = asyncio.get_event_loop()  # pylint: disable=invalid-name
    loop.run_until_complete(_sequential())

    # Wait for any pending logging emits now there's
    # nothing else running anymore
    loop.run_until_complete(asyncio.gather(*logging_future_list))

    # Now that logging is done too, close the connection cleanly.
    loop.run_until_complete(manager_commands.deinit())

    # Any stragglers?
    pending = asyncio.Task.all_tasks()  # pylint: disable=invalid-name
    loop.run_until_complete(asyncio.gather(*pending))

    loop.close()



""".. Ignore pydocstyle D400.

===============
Elastic Indices
===============

Main two classes


.. autoclass:: resolwe.elastic.indices.BaseDocument
    :members:

.. autoclass:: resolwe.elastic.indices.BaseIndex
    :members:

"""
import copy
import logging
import threading

import elasticsearch_dsl as dsl
from elasticsearch.helpers import bulk
from elasticsearch_dsl.connections import connections
from elasticsearch_dsl.exceptions import IllegalOperation

from django.contrib.contenttypes.models import ContentType

from django_priority_batch import PrioritizedBatcher
from guardian.conf.settings import ANONYMOUS_USER_NAME
from guardian.models import GroupObjectPermission, UserObjectPermission

from resolwe.flow.utils import dict_dot

from .composer import composer
from .utils import prepare_connection

__all__ = ('BaseDocument', 'BaseIndex')

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class BaseDocument(dsl.DocType):
    """Base document class to build ElasticSearch documents.

    This is standard ``elasticsearch-dsl`` ``DocType`` class with
    already added fields for handling permissions.

    """

    #: list of user ids with view permission on the object
    users_with_permissions = dsl.Keyword(multi=True)

    #: list of group ids with view permission on the object
    groups_with_permissions = dsl.Keyword(multi=True)

    #: identifies if object has public view permission assigned
    public_permission = dsl.Boolean()


class BaseIndex:
    """Base index class.

    Builds ElasticSearch index for specific type of objects. Index is
    based on document type defined in ``document_type``. Fields are
    determined from document and are populated with one of the following
    methods (in the exact order):

      * ``get_<field_name>_value`` method is used
      * ``mapping[<feild_name>]`` is used - if value is callable, it is
        called with current object as only argument
      * value is extracted from the object's field with the same name

    To make the index, caall ``run`` function. Index is build for all
    objects in queryset. To build index for just one object, specify it
    in ``obj`` parameter of ``run`` function.

    To work properly, subclass of this class must override following
    attributes:

      * object_type - class to which object must belong to be processed
      * document_class - subclass of :class:`BaseDocument` that is used
        to build actual index

    Additional (optional) methods and attributes that can be overriden
    are:

      * mapping - mapping for transforming object into index
      * :func:`~BaseIndex.preprocess_object`
      * :func:`~BaseIndex.filter`

    """

    #: queryset of objects to index
    queryset = None

    #: type of object that are indexed, i.e. Django model
    object_type = None

    #: document class used to create index
    document_class = None

    #: auto generated ES index postfix used in tests
    testing_postfix = ''

    #: mapping used for building document
    mapping = {}

    def __init__(self):
        """Perform initial checks and save given object."""
        class_name = type(self).__name__
        if not self.object_type:
            raise RuntimeError('`object_type` must be defined in {}'.format(class_name))

        if not self.document_class:
            raise RuntimeError('`document_class` must be defined in {}'.format(class_name))

        if self.queryset is None:
            raise RuntimeError('`queryset` must be defined in {}'.format(class_name))

        # Apply any extensions defined for the given document class. Document class extensions
        # come in the form of field definitions.
        self.document_class = copy.deepcopy(self.document_class)
        for extension in composer.get_extensions(self.document_class):
            for name in dir(extension):
                field = getattr(extension, name)
                if isinstance(field, dsl.Field):
                    self.document_class._doc_type.mapping.field(name, field)  # pylint: disable=protected-access

        #: list of built documents waiting to be pushed
        self.push_queue = []

        self._index_name = self.document_class()._get_index()  # pylint: disable=not-callable,protected-access
        self._mapping_created = False

        #: id of thread id where connection was established
        self.connection_thread_id = None

    def _refresh_connection(self):
        """Refresh connection to Elasticsearch when worker is started.

        File descriptors (sockets) can be shared between multiple
        threads. If same connection is used by multiple threads at the
        same time, this can cause timeouts in some of the pushes. So
        connection needs to be reestablished in each thread to make sure
        that it is unique per thread.
        """
        # Thread with same id can be created when one terminates, but it
        # is ok, as we are only concerned about concurent pushes.
        current_thread_id = threading.current_thread().ident

        if current_thread_id != self.connection_thread_id:
            prepare_connection()

            self.connection_thread_id = current_thread_id

    def filter(self, obj):
        """Determine if object should be processed.

        If ``False`` is returned, processingg of the current object will
        be aborted.
        """
        pass

    def preprocess_object(self, obj):
        """Preprocess object before indexing.

        This function is called before `func:process_object`. It can be
        used for advanced pre-processing of the object, i.e. adding
        annotations that will be used in multiple fields.
        """
        return obj

    def get_object_id(self, obj):
        """Return unique identifier of the object.

        Object's id is returned by default. This method can be overriden
        if object doesn't have ``id`` attribute.
        """
        return obj.id

    def generate_id(self, obj):
        """Generate unique document id for ElasticSearch."""
        object_type = type(obj).__name__.lower()
        return '{}_{}'.format(object_type, self.get_object_id(obj))

    def process_object(self, obj):
        """Process current object and push it to the ElasticSearch."""
        document = self.document_class(meta={'id': self.generate_id(obj)})  # pylint: disable=not-callable

        for field in document._doc_type.mapping:  # pylint: disable=protected-access
            if field in ['users_with_permissions', 'groups_with_permissions', 'public_permission']:
                continue  # These fields are handled separately

            try:
                # use get_X_value function
                get_value_function = getattr(self, 'get_{}_value'.format(field), None)
                if get_value_function:
                    setattr(document, field, get_value_function(obj))  # pylint: disable=not-callable
                    continue

                # use `mapping` dict
                if field in self.mapping:
                    if callable(self.mapping[field]):
                        setattr(document, field, self.mapping[field](obj))
                        continue

                    try:
                        object_attr = dict_dot(obj, self.mapping[field])
                    except (KeyError, AttributeError):
                        object_attr = None

                    if callable(object_attr):
                        # use method on object
                        setattr(document, field, object_attr(obj))
                    else:
                        # use attribute on object
                        setattr(document, field, object_attr)
                    continue

                # get value from the object
                try:
                    object_value = dict_dot(obj, field)
                    setattr(document, field, object_value)
                    continue
                except KeyError:
                    pass

                raise AttributeError("Cannot determine mapping for field {}".format(field))

            except:  # noqa pylint: disable=bare-except
                logger.exception(
                    "Error occurred while setting value of field '%s' in '%s' Elasticsearch index.",
                    field, self.__class__.__name__,
                    extra={'object_type': self.object_type, 'obj_id': obj.pk}
                )

        permissions = self.get_permissions(obj)
        document.users_with_permissions = permissions['users']
        document.groups_with_permissions = permissions['groups']
        document.public_permission = permissions['public']

        self.push_queue.append(document)

    def create_mapping(self):
        """Create the mappings in elasticsearch."""
        try:
            self.document_class.init()
            self._mapping_created = True
        except IllegalOperation as error:
            if error.args[0].startswith('You cannot update analysis configuration'):
                # Ignore mapping update errors, which are thrown even when the analysis
                # configuration stays the same.
                # TODO: Remove this when https://github.com/elastic/elasticsearch-dsl-py/pull/272 is merged.
                return

            raise

    def build(self, obj=None, queryset=None, push=True):
        """Build indexes."""
        if obj is not None and queryset is not None:
            raise ValueError(
                "Only one of 'obj' and 'queryset' parameters can be passed to the build method."
            )

        if obj is not None:
            if self.queryset.model != obj._meta.model:  # pylint: disable=protected-access
                logger.debug(
                    "Object type mismatch, skipping build of '%s' Elasticsearch index.",
                    self.__class__.__name__
                )
                return

            if not self.queryset.filter(pk=self.get_object_id(obj)).exists():
                logger.debug(
                    "Object not in predefined queryset, skipping build of '%s' Elasticsearch index.",
                    self.__class__.__name__
                )
                return

        elif queryset is not None:
            if self.queryset.model != queryset.model:
                logger.debug(
                    "Queryset type mismatch, skipping build of '%s' Elasticsearch index.",
                    self.__class__.__name__
                )
                return

        FULL_REBUILD = 'full'  # pylint: disable=invalid-name

        def handler(agg=None):
            """Index build handler."""
            if agg == FULL_REBUILD:
                queryset = self.queryset.all()
            else:
                queryset = self.queryset.none().union(*agg)

            self._build(queryset=queryset, push=push)

        def aggregator(agg=None):
            """Index build aggregator."""
            if agg == FULL_REBUILD:
                # A full rebuild is required, ignore any other builds.
                pass
            else:
                if agg is None:
                    agg = []

                if obj is not None:
                    # Build of a single object.
                    agg.append(self.queryset.filter(pk=obj.pk))
                elif queryset is not None:
                    # Build of multiple objects.
                    agg.append(queryset)
                else:
                    # Full rebuild, ignore any other builds.
                    agg = FULL_REBUILD

            return agg

        batcher = PrioritizedBatcher.global_instance()
        if batcher.is_started:
            batcher.add('resolwe.elastic', handler, group_by=(self._index_name, push), aggregator=aggregator)
        else:
            self._build(obj=obj, queryset=queryset, push=push)

    def _build(self, obj=None, queryset=None, push=True):
        """Build indexes."""
        logger.debug("Building '%s' Elasticsearch index...", self.__class__.__name__)

        if obj is not None:
            build_list = [obj]

        elif queryset is not None:
            build_list = self.queryset.intersection(queryset)

            logger.debug("Found %s elements to build.", build_list.count())

        else:
            build_list = self.queryset.all()

            logger.debug("Found %s elements to build.", build_list.count())

        for obj in build_list:
            if self.filter(obj) is False:
                continue

            try:
                obj = self.preprocess_object(obj)
            except:  # noqa pylint: disable=bare-except
                logger.exception(
                    "Error occurred while preprocessing '%s' Elasticsearch index.",
                    self.__class__.__name__,
                    extra={'object_type': self.object_type, 'obj_id': obj.pk}
                )

            try:
                self.process_object(obj)
            except:  # noqa pylint: disable=bare-except
                logger.exception(
                    "Error occurred while processing '%s' Elasticsearch index.",
                    self.__class__.__name__,
                    extra={'object_type': self.object_type, 'obj_id': obj.pk}
                )

        logger.debug("Finished building '%s' Elasticsearch index.", self.__class__.__name__)

        if push:
            self.push()

    def push(self):
        """Push built documents to ElasticSearch."""
        self._refresh_connection()

        # Check if we need to update mappings as this needs to be done
        # before we push anything to the Elasticsearch server.
        # This must be done even if the queue is empty, as otherwise ES
        # will fail when retrieving data.
        if not self._mapping_created:
            logger.debug("Pushing mapping for Elasticsearch index '%s'.", self.__class__.__name__)
            self.create_mapping()

        if not self.push_queue:
            logger.debug("No documents to push, skipping push.")
            return

        logger.debug("Found %s documents to push to Elasticsearch.", len(self.push_queue))

        bulk(connections.get_connection(), (doc.to_dict(True) for doc in self.push_queue), refresh=True)
        self.push_queue = []

        logger.debug("Finished pushing builded documents to Elasticsearch server.")

    def destroy(self):
        """Destroy an index."""
        self._refresh_connection()

        self.push_queue = []
        index_name = self.document_class()._get_index()  # pylint: disable=protected-access,not-callable
        connections.get_connection().indices.delete(index_name, ignore=404)  # pylint: disable=no-member

        self._mapping_created = False

    def get_permissions(self, obj):
        """Return users and groups with ``view`` permission on the current object.

        Return a dict with two keys - ``users`` and ``groups`` - which
        contain list of ids of users/groups with ``view`` permission.
        """
        # TODO: Optimize this for bulk running
        filters = {
            'object_pk': obj.id,
            'content_type': ContentType.objects.get_for_model(obj),
            'permission__codename__startswith': 'view',
        }
        return {
            'users': list(
                UserObjectPermission.objects.filter(**filters).distinct('user').values_list('user_id', flat=True)
            ),
            'groups': list(
                GroupObjectPermission.objects.filter(**filters).distinct('group').values_list('group', flat=True)
            ),
            'public': UserObjectPermission.objects.filter(user__username=ANONYMOUS_USER_NAME, **filters).exists(),
        }

    def get_dependencies(self):
        """Return dependencies, which should trigger updates of this index."""
        return []

    def remove_object(self, obj):
        """Remove current object from the ElasticSearch."""
        obj_id = self.generate_id(obj)
        es_obj = self.document_class.get(obj_id, ignore=[404])
        # Object may not exist in this index.
        if es_obj:
            es_obj.delete(refresh=True)

    def search(self):
        """Return search query of document object."""
        return self.document_class.search()



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-03-26 08:39
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('flow', '0013_migration_history'),
    ]

    operations = [
        migrations.AddField(
            model_name='data',
            name='process_cores',
            field=models.PositiveSmallIntegerField(default=0),
        ),
        migrations.AddField(
            model_name='data',
            name='process_memory',
            field=models.PositiveIntegerField(default=0),
        ),
    ]



"""Resolwe custom serializer fields."""
from django.core.exceptions import ObjectDoesNotExist
from django.utils.encoding import smart_text

from rest_framework.relations import RelatedField

from resolwe.flow.models import DescriptorSchema
from resolwe.permissions.shortcuts import get_objects_for_user
from resolwe.permissions.utils import get_full_perm

from .descriptor import DescriptorSchemaSerializer


class ResolweSlugRelatedField(RelatedField):
    """
    Resolwe specific implementation of SlugRelatedField.

    A read-write field that represents the target of the relationship
    by a unique combination of 'slug' and 'version' attributes.

    This is a modification of rest_framework.relations.SlugRelatedField.
    Since slug is not unique, (but combination of slug and version is),
    we filter objects by slug, by permissions and return object with
    highest version.

    """

    default_error_messages = {
        'does_not_exist': ('Invalid {model_name} {slug_name} "{value}" - object does not exist.'),
        'invalid': ('Invalid value.'),
    }

    def __init__(self, slug_field='slug', **kwargs):
        """Initialize attributes."""
        self.slug_field = slug_field
        super().__init__(**kwargs)

    def to_internal_value(self, data):
        """Convert to internal value."""
        user = getattr(self.context.get('request'), 'user')
        queryset = self.get_queryset()
        permission = get_full_perm('view', queryset.model)
        try:
            return get_objects_for_user(
                user,
                permission,
                queryset.filter(**{self.slug_field: data}),
            ).latest()
        except ObjectDoesNotExist:
            self.fail(
                'does_not_exist',
                slug_name=self.slug_field,
                value=smart_text(data),
                model_name=queryset.model._meta.model_name,  # pylint: disable=protected-access
            )
        except (TypeError, ValueError):
            self.fail('invalid')

    def to_representation(self, obj):
        """Convert to representation."""
        return obj.pk


class NestedDescriptorSchemaSerializer(ResolweSlugRelatedField):
    """DescriptorSchema specific implementation of ResolweSlugRelatedField."""

    def __init__(self, **kwargs):
        """Initialize attributes."""
        kwargs['queryset'] = DescriptorSchema.objects.all()
        super().__init__(slug_field='slug', **kwargs)

    def to_representation(self, obj):
        """Convert to representation."""
        return DescriptorSchemaSerializer(obj, required=self.required).data



# pylint: disable=missing-docstring,too-many-lines
import io
import json
import os
import shutil
from datetime import datetime, timedelta

from mock import MagicMock, patch

from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError
from django.db import transaction

from guardian.shortcuts import assign_perm, get_perms, remove_perm
from rest_framework.test import APIRequestFactory, APITestCase, force_authenticate

from resolwe.flow.expression_engines import EvaluationError
from resolwe.flow.models import Collection, DataMigrationHistory, DescriptorSchema, Entity, Process, Storage
from resolwe.flow.models.data import Data, DataDependency, DataLocation, hydrate_size, render_template
from resolwe.flow.models.utils import hydrate_input_references
from resolwe.flow.views import DataViewSet
from resolwe.test import TestCase, TransactionTestCase

try:
    import builtins  # py3
except ImportError:
    import __builtin__ as builtins  # py2


class DataModelNameTest(TransactionTestCase):

    def test_name(self):
        process = Process.objects.create(slug='test-first',
                                         type='data:test:first:',
                                         contributor=self.contributor,
                                         data_name='Process based data name',
                                         output_schema=[{
                                             'name': 'stat',
                                             'type': 'basic:string:',
                                             'required': False,
                                         }],
                                         run={
                                             'language': 'bash',
                                             'program': 'echo {"stat": "42"}'
                                         })

        data = Data.objects.create(contributor=self.contributor,
                                   process=process)
        data.refresh_from_db()

        self.assertEqual(data.name, 'Process based data name')
        self.assertFalse(data.named_by_user)

        data.name = 'Name changed by user'
        data.save()
        data.refresh_from_db()

        self.assertEqual(data.name, 'Name changed by user')
        self.assertTrue(data.named_by_user)

        data = Data.objects.create(name='Explicit data name',
                                   contributor=self.contributor,
                                   process=process)
        data.refresh_from_db()

        self.assertEqual(data.name, 'Explicit data name')
        self.assertTrue(data.named_by_user)

        process = Process.objects.create(slug='test-second',
                                         type='test:second',
                                         contributor=self.contributor,
                                         requirements={'expression-engine': 'jinja'},
                                         data_name='Process based data name, value: {{src.stat}}',
                                         input_schema=[{
                                             'name': 'src',
                                             'type': 'data:test:first:',
                                             'required': False,
                                         }])

        with transaction.atomic():
            second = Data.objects.create(contributor=self.contributor,
                                         process=process,
                                         input={'src': data.id})

            data.output = {'stat': '42'}
            data.status = 'OK'
            data.save()

            self.assertEqual(second.name, 'Process based data name, value: ')
            self.assertFalse(second.named_by_user)

        second = Data.objects.get(id=second.id)

        self.assertEqual(second.name, 'Process based data name, value: 42')
        self.assertFalse(second.named_by_user)

        with transaction.atomic():
            data.output = {}
            data.status = 'RE'
            data.save()

            second = Data.objects.create(contributor=self.contributor,
                                         process=process,
                                         input={'src': data.id})

            second.name = 'User\' data name'
            second.save()

            data.output = {'stat': '42'}
            data.status = 'OK'
            data.save()

            self.assertEqual(second.name, 'User\' data name')
            self.assertTrue(second.named_by_user)

        second = Data.objects.get(id=second.id)

        self.assertEqual(second.name, 'User\' data name')
        self.assertTrue(second.named_by_user)


class DataModelTest(TestCase):

    def test_delete_chunked(self):
        process = Process.objects.create(contributor=self.contributor)
        for _ in range(1000):
            Data.objects.create(contributor=self.contributor, process=process)

        Data.objects.all().delete_chunked(chunk_size=50)
        self.assertFalse(Data.objects.exists())

    def test_trim_name(self):
        process = Process.objects.create(contributor=self.contributor, data_name='test' * 50)
        data = Data.objects.create(contributor=self.contributor, process=process)

        self.assertEqual(len(data.name), 100)
        self.assertEqual(data.name[-3:], '...')

    def test_hydrate_file_size(self):
        proc = Process.objects.create(
            name='Test process',
            contributor=self.contributor,
            output_schema=[
                {'name': 'output_file', 'type': 'basic:file:'}
            ]
        )

        data = Data.objects.create(
            name='Test data',
            contributor=self.contributor,
            process=proc,
        )

        data.output = {
            'output_file': {'file': 'output.txt'}
        }

        data_location = DataLocation.objects.create(subpath='')
        data_location.subpath = str(data_location.id)
        data_location.save()
        data_location.data.add(data)

        dir_path = data.location.get_path()
        os.makedirs(dir_path)

        with self.assertRaises(ValidationError):
            data.save()

        file_path = os.path.join(dir_path, 'output.txt')
        with open(file_path, 'w') as fn:
            fn.write('foo bar')

        data.save()

        self.assertEqual(data.output['output_file']['size'], 7)

    def test_dependencies_single(self):
        process = Process.objects.create(slug='test-dependencies',
                                         type='data:test:dependencies:',
                                         contributor=self.contributor,
                                         input_schema=[{
                                             'name': 'src',
                                             'type': 'data:test:dependencies:',
                                             'required': False,
                                         }])

        first = Data.objects.create(name='First data',
                                    contributor=self.contributor,
                                    process=process)

        second = Data.objects.create(name='Second data',
                                     contributor=self.contributor,
                                     process=process,
                                     input={'src': first.id})

        third = Data.objects.create(name='Third data',
                                    contributor=self.contributor,
                                    process=process,
                                    input={'src': first.id})

        self.assertEqual(first.parents.all().count(), 0)
        self.assertEqual(first.children.all().count(), 2)
        self.assertEqual(second.children.all().count(), 0)
        self.assertEqual(second.parents.all().count(), 1)
        self.assertEqual(third.children.all().count(), 0)
        self.assertEqual(third.parents.all().count(), 1)
        self.assertIn(first, second.parents.all())
        self.assertIn(first, third.parents.all())
        self.assertIn(second, first.children.all())
        self.assertIn(third, first.children.all())

        # Check correct dependency type is created.
        self.assertEqual({d.kind for d in first.children_dependency.all()}, {DataDependency.KIND_IO})
        self.assertEqual({d.kind for d in second.parents_dependency.all()}, {DataDependency.KIND_IO})
        self.assertEqual({d.kind for d in third.parents_dependency.all()}, {DataDependency.KIND_IO})

    def test_dependencies_list(self):
        process = Process.objects.create(slug='test-dependencies-list',
                                         type='data:test:dependencies:list:',
                                         contributor=self.contributor,
                                         input_schema=[{
                                             'name': 'src',
                                             'type': 'list:data:test:dependencies:list:',
                                             'required': False,
                                         }])

        first = Data.objects.create(name='First data',
                                    contributor=self.contributor,
                                    process=process)

        second = Data.objects.create(name='Second data',
                                     contributor=self.contributor,
                                     process=process,
                                     input={'src': [first.id]})

        third = Data.objects.create(name='Third data',
                                    contributor=self.contributor,
                                    process=process,
                                    input={'src': [first.id, second.id]})

        self.assertEqual(first.parents.all().count(), 0)
        self.assertEqual(first.children.all().count(), 2)
        self.assertEqual(second.children.all().count(), 1)
        self.assertEqual(second.parents.all().count(), 1)
        self.assertEqual(third.children.all().count(), 0)
        self.assertEqual(third.parents.all().count(), 2)
        self.assertIn(first, second.parents.all())
        self.assertIn(first, third.parents.all())
        self.assertIn(second, first.children.all())
        self.assertIn(third, first.children.all())
        self.assertIn(third, second.children.all())

        # Check correct dependency type is created.
        self.assertEqual({d.kind for d in first.children_dependency.all()}, {DataDependency.KIND_IO})
        self.assertEqual({d.kind for d in second.children_dependency.all()}, {DataDependency.KIND_IO})
        self.assertEqual({d.kind for d in second.parents_dependency.all()}, {DataDependency.KIND_IO})
        self.assertEqual({d.kind for d in third.parents_dependency.all()}, {DataDependency.KIND_IO})


class EntityModelTest(TestCase):

    def setUp(self):
        super().setUp()

        DescriptorSchema.objects.create(name='Sample', slug='sample', contributor=self.contributor)
        self.process = Process.objects.create(name='Test process',
                                              type='data:test:',
                                              contributor=self.contributor,
                                              entity_type='sample',
                                              entity_descriptor_schema='sample')
        # Entity is created automatically when Data object is created
        self.data = Data.objects.create(name='Test data', contributor=self.contributor, process=self.process)

    def test_delete_data(self):
        # Create another Data object and add it to the same Entity.
        data_2 = Data.objects.create(name='Test data', contributor=self.contributor, process=self.process)
        data_2.entity_set.all().delete()
        self.data.entity_set.first().data.add(data_2)

        data_2.delete()
        self.assertEqual(Entity.objects.count(), 1)

        self.data.delete()
        self.assertEqual(Entity.objects.count(), 0)

    def test_two_entities(self):
        entity_2 = Entity.objects.create(contributor=self.contributor)
        entity_2.data.add(self.data)

        self.data.delete()
        self.assertEqual(Entity.objects.count(), 0)

    def test_new_sample(self):
        data = Data.objects.create(name='Test data', contributor=self.contributor, process=self.process,
                                   tags=['foo', 'bar'])
        entity = Entity.objects.last()
        self.assertTrue(entity.data.filter(pk=data.pk).exists())

        # Make sure tags are copied.
        self.assertEqual(entity.tags, data.tags)

    def test_entity_inheritance(self):
        # Prepare Data objects with entities.
        process_entity = self.process
        data_1 = self.data
        data_2 = Data.objects.create(contributor=self.contributor, process=process_entity)
        # Prepare Data objects without entities.
        process_no_entity = Process.objects.create(type='data:test:', contributor=self.contributor)
        data_3 = Data.objects.create(contributor=self.contributor, process=process_no_entity)
        # Prepare test process.
        test_process = Process.objects.create(
            contributor=self.contributor,
            entity_type='sample',
            entity_descriptor_schema='sample',
            input_schema=[
                {'name': 'data_list', 'type': 'list:data:test:', 'required': False},
                {'name': 'data', 'type': 'data:test:', 'required': False},
            ],
        )

        # Single Entity - Data object should be added to it.
        data = Data.objects.create(
            contributor=self.contributor,
            process=test_process,
            input={
                'data_list': [data_1.pk, data_3.pk],
            }
        )
        self.assertEqual(data.entity_set.first(), data_1.entity_set.first())

        # Multiple Entities - Data object should be added to none of them.
        data = Data.objects.create(
            contributor=self.contributor,
            process=test_process,
            input={
                'data_list': [data_1.pk, data_2.pk],
            }
        )
        self.assertEqual(data.entity_set.count(), 0)

        # Multiple Entities with entity_input defined - Data object should be added to it.
        test_process.entity_input = 'data'
        test_process.save()
        data = Data.objects.create(
            contributor=self.contributor,
            process=test_process,
            input={
                'data': data_1.pk,
                'data_list': [data_2.pk],
            }
        )
        self.assertEqual(data.entity_set.first(), data_1.entity_set.first())
        test_process.entity_input = 'data_list'
        test_process.save()
        data = Data.objects.create(
            contributor=self.contributor,
            process=test_process,
            input={
                'data': data_1.pk,
                'data_list': [data_2.pk],
            }
        )
        self.assertEqual(data.entity_set.first(), data_2.entity_set.first())
        test_process.entity_input = None
        test_process.save()

        # Entities of different types - Data object should be added to the right one.
        entity_2 = data_2.entity_set.first()
        entity_2.type = 'something_else'
        entity_2.save()
        data = Data.objects.create(
            contributor=self.contributor,
            process=test_process,
            input={
                'data_list': [data_1.pk, data_2.pk],
            }
        )
        self.assertEqual(data.entity_set.first(), data_1.entity_set.first())
        entity_2.type = 'sample'
        entity_2.save()


class GetOrCreateTestCase(APITestCase):

    def setUp(self):
        super().setUp()

        user_model = get_user_model()
        self.user = user_model.objects.create(username='test_user', password='test_pwd')

        self.process = Process.objects.create(
            name='Temporary process',
            contributor=self.user,
            slug='tmp-process',
            persistence=Process.PERSISTENCE_TEMP,
            input_schema=[
                {'name': 'some_value', 'type': 'basic:integer:', 'default': 42}
            ],
        )
        assign_perm('view_process', self.user, self.process)

        process_2 = Process.objects.create(
            name='Another process',
            contributor=self.user,
            slug='another-process',
            persistence=Process.PERSISTENCE_TEMP,
            input_schema=[
                {'name': 'some_value', 'type': 'basic:integer:'}
            ],
        )
        assign_perm('view_process', self.user, process_2)

        self.data = Data.objects.create(
            name='Temporary data',
            contributor=self.user,
            process=self.process,
            input={'some_value': 42}
        )
        assign_perm('view_data', self.user, self.data)

        self.get_or_create_view = DataViewSet.as_view({'post': 'get_or_create'})
        self.factory = APIRequestFactory()

    def tearDown(self):
        for data in Data.objects.all():
            if data.location:
                data_dir = data.location.get_path()
                shutil.rmtree(data_dir, ignore_errors=True)

        super().tearDown()

    def test_get_same(self):
        request = self.factory.post(
            '',
            {'name': 'Data object', 'input': {'some_value': 42}, 'process': 'tmp-process'},
            format='json'
        )
        force_authenticate(request, user=self.user)

        response = self.get_or_create_view(request)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['id'], self.data.pk)

    def test_use_defaults(self):
        request = self.factory.post(
            '',
            {'name': 'Data object', 'input': {}, 'process': 'tmp-process'},
            format='json'
        )
        force_authenticate(request, user=self.user)

        response = self.get_or_create_view(request)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['id'], self.data.pk)

    def test_missing_permission(self):
        remove_perm('view_data', self.user, self.data)

        request = self.factory.post(
            '',
            {'name': 'Data object', 'input': {'some_value': 42}, 'process': 'tmp-process'},
            format='json'
        )
        force_authenticate(request, user=self.user)

        response = self.get_or_create_view(request)
        self.assertEqual(response.status_code, 201)
        self.assertNotEqual(response.data['id'], self.data.pk)

    def test_different_input(self):
        request = self.factory.post(
            '',
            {'name': 'Data object', 'input': {'some_value': 43}, 'process': 'tmp-process'},
            format='json'
        )
        force_authenticate(request, user=self.user)

        response = self.get_or_create_view(request)
        self.assertEqual(response.status_code, 201)
        self.assertNotEqual(response.data['id'], self.data.pk)

    def test_different_process(self):
        request = self.factory.post(
            '',
            {'name': 'Data object', 'input': {'some_value': 43}, 'process': 'another-process'},
            format='json'
        )
        force_authenticate(request, user=self.user)

        response = self.get_or_create_view(request)
        self.assertEqual(response.status_code, 201)
        self.assertNotEqual(response.data['id'], self.data.pk)

    def test_different_process_version(self):
        self.process.version = '2.0.0'
        self.process.save()

        request = self.factory.post(
            '',
            {'name': 'Data object', 'input': {'some_value': 42}, 'process': 'tmp-process'},
            format='json'
        )
        force_authenticate(request, user=self.user)

        response = self.get_or_create_view(request)
        self.assertEqual(response.status_code, 201)
        self.assertNotEqual(response.data['id'], self.data.pk)

    def test_raw_process(self):
        self.process.persistence = Process.PERSISTENCE_RAW
        self.process.save()

        request = self.factory.post(
            '',
            {'name': 'Data object', 'input': {'some_value': 42}, 'process': 'tmp-process'},
            format='json'
        )
        force_authenticate(request, user=self.user)

        response = self.get_or_create_view(request)
        self.assertEqual(response.status_code, 201)
        self.assertNotEqual(response.data['id'], self.data.pk)


class DuplicateTestCase(TestCase):

    def test_data_duplicate(self):
        process1 = Process.objects.create(
            contributor=self.user,
            type='data:test:first:',
        )
        process2 = Process.objects.create(
            contributor=self.user,
            input_schema=[
                {'name': 'data_field', 'type': 'data:test:first:'},
            ],
            output_schema=[
                {'name': 'json_field', 'type': 'basic:json:'},
            ],
        )

        input_data = Data.objects.create(
            name='Data 1',
            contributor=self.user,
            process=process1,
        )
        input_data.status = Data.STATUS_DONE
        input_data.save()

        data2 = Data.objects.create(
            name='Data 2',
            contributor=self.user,
            process=process2,
            started=datetime.now(),
            input={'data_field': input_data.id},
        )
        data_location = DataLocation.objects.create(subpath='')
        data_location.subpath = str(data_location.id)
        data_location.save()
        data_location.data.add(data2)
        data2.output = {'json_field': {'foo': 'bar'}}
        data2.status = Data.STATUS_DONE
        data2.save()
        data2.migration_history.create(migration='migration_1')
        data2.migration_history.create(migration='migration_2')

        # Duplicate.
        duplicates = Data.objects.filter(id=data2.id).duplicate(self.contributor)
        self.assertEqual(len(duplicates), 1)
        duplicate = duplicates[0]
        self.assertTrue(duplicate.is_duplicate())

        # Convert original and duplicated to dict.
        data2_dict = Data.objects.filter(id=data2.id).values()[0]
        duplicate_dict = Data.objects.filter(id=duplicate.id).values()[0]

        # Pop fields that should differ and assert the remaining.
        fields_to_differ = ('id', 'slug', 'contributor_id', 'name', 'duplicated', 'modified')
        for model_dict in (data2_dict, duplicate_dict):
            for field in fields_to_differ:
                model_dict.pop(field)

        self.assertDictEqual(data2_dict, duplicate_dict)

        # Assert location.
        self.assertEqual(data2.location.id, duplicate.location.id)

        # Assert fields that differ.
        self.assertEqual(duplicate.slug, 'copy-of-data-2')
        self.assertEqual(duplicate.name, 'Copy of Data 2')
        self.assertEqual(duplicate.contributor.username, 'contributor')
        self.assertAlmostEqual(duplicate.duplicated, datetime.now(), delta=timedelta(seconds=3))
        self.assertAlmostEqual(duplicate.modified, datetime.now(), delta=timedelta(seconds=3))

        # Assert dependencies.
        self.assertEqual(DataDependency.objects.count(), 2)
        self.assertTrue(
            DataDependency.objects.filter(
                kind=DataDependency.KIND_IO,
                parent=input_data,
                child=data2
            ).exists()
        )
        self.assertTrue(
            DataDependency.objects.filter(
                kind=DataDependency.KIND_IO,
                parent=input_data,
                child=duplicate
            ).exists()
        )

        # Assert storage
        self.assertEqual(Storage.objects.count(), 1)
        self.assertEqual(data2.storages.first(), Storage.objects.first())
        self.assertEqual(duplicate.storages.first(), Storage.objects.first())

        # Assert migration history
        self.assertEqual(DataMigrationHistory.objects.count(), 4)
        self.assertEqual(data2.migration_history.earliest('created').migration, 'migration_1')
        self.assertEqual(data2.migration_history.latest('created').migration, 'migration_2')
        self.assertEqual(duplicate.migration_history.earliest('created').migration, 'migration_1')
        self.assertEqual(duplicate.migration_history.latest('created').migration, 'migration_2')

        # Assert permissions.
        self.assertEqual(len(get_perms(self.contributor, duplicate)), 5)

    def test_data_duplicate_duplicate(self):
        process1 = Process.objects.create(
            contributor=self.user,
            type='data:test:first:',
        )
        process2 = Process.objects.create(
            contributor=self.user,
            input_schema=[
                {'name': 'data_field', 'type': 'data:test:first:'},
            ],
            output_schema=[
                {'name': 'json_field', 'type': 'basic:json:'},
            ],
        )

        input_data = Data.objects.create(contributor=self.user, process=process1)

        data = Data.objects.create(contributor=self.user, process=process2, input={'data_field': input_data.id})
        data_location = DataLocation.objects.create(subpath='')
        data_location.subpath = str(data_location.id)
        data_location.save()
        data_location.data.add(data)
        data.output = {'json_field': {'foo': 'bar'}}
        data.status = Data.STATUS_DONE
        data.save()
        data.migration_history.create(migration='migration_1')

        # Duplicate.
        duplicate = data.duplicate(self.contributor)
        duplicate_of_duplicate = duplicate.duplicate()

        self.assertEqual(duplicate.contributor.id, duplicate_of_duplicate.contributor.id)

        # Assert location
        self.assertEqual(duplicate.location.id, duplicate_of_duplicate.location.id)

        # Assert dependencies.
        self.assertEqual(DataDependency.objects.count(), 3)
        self.assertTrue(
            DataDependency.objects.filter(
                kind=DataDependency.KIND_IO,
                parent=input_data,
                child=duplicate_of_duplicate
            ).exists()
        )

        # Assert storage
        self.assertEqual(duplicate_of_duplicate.storages.first().id, Storage.objects.first().id)

        # Assert migration history
        self.assertEqual(DataMigrationHistory.objects.count(), 3)
        self.assertEqual(duplicate_of_duplicate.migration_history.latest('created').migration, 'migration_1')

        # Assert permissions.
        self.assertEqual(len(get_perms(self.contributor, duplicate_of_duplicate)), 5)

    def test_input_rewiring(self):
        process1 = Process.objects.create(
            contributor=self.user,
            type='data:test:first:',
        )
        process2 = Process.objects.create(
            contributor=self.user,
            type='data:test:second:',
            input_schema=[
                {'name': 'data_field1', 'type': 'data:test:first:'},
                {'name': 'data_field2', 'type': 'data:test:first:'},
                {'name': 'data_list_field', 'type': 'list:data:test:first:'},
            ],
        )

        data1 = Data.objects.create(
            contributor=self.user,
            process=process1,
            status=Data.STATUS_DONE,
        )

        should_not_be_rewritten = data1.duplicate()

        data2 = Data.objects.create(
            contributor=self.user,
            process=process2,
            input={
                'data_field1': data1.id,
                'data_field2': should_not_be_rewritten.id,
                'data_list_field': [data1.id, should_not_be_rewritten.id],
            },
            status=Data.STATUS_DONE,
        )

        # Duplicate.
        duplicates = Data.objects.filter(id__in=[data1.id, data2.id]).duplicate(self.contributor)
        duplicate1, duplicate2 = duplicates

        self.assertEqual(duplicate2.input['data_field1'], duplicate1.id)
        self.assertEqual(duplicate2.input['data_field2'], should_not_be_rewritten.id)
        self.assertEqual(duplicate2.input['data_list_field'], [duplicate1.id, should_not_be_rewritten.id])

    def test_data_status_not_done(self):
        process = Process.objects.create(contributor=self.user)
        data = Data.objects.create(contributor=self.user, process=process, status=Data.STATUS_WAITING)
        data2 = Data.objects.create(contributor=self.user, process=process, status=Data.STATUS_DONE)

        with self.assertRaisesMessage(ValidationError, 'Data object must have done or error status to be duplicated'):
            Data.objects.filter(id__in=[data.id, data2.id]).duplicate(self.contributor)

    def test_data_long_name(self):
        process = Process.objects.create(contributor=self.user)

        long_name = 'a' * (Data._meta.get_field('name').max_length - 1)  # pylint: disable=protected-access
        data = Data.objects.create(name=long_name, contributor=self.user, process=process, status=Data.STATUS_DONE)

        duplicate = Data.objects.filter(id=data.id).duplicate(self.contributor)[0]

        self.assertTrue(duplicate.name.startswith('Copy of '))
        self.assertTrue(duplicate.name.endswith('...'))
        self.assertEqual(
            len(duplicate.name), Data._meta.get_field('name').max_length  # pylint: disable=protected-access
        )

    def test_entity_duplicate(self):
        process = Process.objects.create(contributor=self.user)
        data = Data.objects.create(name='Data 1', contributor=self.user, process=process, status=Data.STATUS_DONE)
        assign_perm('view_data', self.contributor, data)
        data2 = Data.objects.create(name='Data 2', contributor=self.user, process=process, status=Data.STATUS_DONE)
        assign_perm('view_data', self.contributor, data2)
        entity = Entity.objects.create(name='Entity', contributor=self.user)
        entity.data.add(data, data2)

        # Add to collection.
        collection = Collection.objects.create(contributor=self.user)
        entity.collections.add(collection.id)
        collection.data.add(data, data2)

        # Duplicate.
        entities = Entity.objects.filter(id=entity.id).duplicate(self.contributor)
        self.assertEqual(len(entities), 1)
        duplicate = entities[0]

        entity_dict = Entity.objects.filter(id=entity.id).values()[0]
        duplicate_dict = Entity.objects.filter(id=duplicate.id).values()[0]

        data1_dict = Data.objects.filter(id=data.id).values()[0]
        data1_duplicate = duplicate.data.first()
        data1_duplicate_dict = Data.objects.filter(id=data1_duplicate.id).values()[0]

        data2_dict = Data.objects.filter(id=data2.id).values()[0]
        data2_duplicate = duplicate.data.last()
        data2_duplicate_dict = Data.objects.filter(id=data2_duplicate.id).values()[0]

        self.assertTrue(duplicate.is_duplicate())
        self.assertTrue(data1_duplicate.is_duplicate())

        # Pop fields that should differ and assert the remaining.
        fields_to_differ = ('id', 'slug', 'contributor_id', 'name', 'duplicated', 'modified')
        for model_dict in (
                entity_dict, duplicate_dict, data1_dict, data1_duplicate_dict, data2_dict, data2_duplicate_dict
        ):
            for field in fields_to_differ:
                model_dict.pop(field)

        self.assertDictEqual(entity_dict, duplicate_dict)
        self.assertDictEqual(data1_dict, data1_duplicate_dict)
        self.assertDictEqual(data2_dict, data2_duplicate_dict)

        # Assert fields that differ.
        self.assertEqual(duplicate.slug, 'copy-of-entity')
        self.assertEqual(duplicate.name, 'Copy of Entity')
        self.assertEqual(duplicate.contributor.username, 'contributor')
        self.assertAlmostEqual(duplicate.duplicated, datetime.now(), delta=timedelta(seconds=3))
        self.assertAlmostEqual(duplicate.modified, datetime.now(), delta=timedelta(seconds=3))

        self.assertEqual(data1_duplicate.slug, 'copy-of-data-1')
        self.assertEqual(data1_duplicate.name, 'Copy of Data 1')
        self.assertEqual(data2_duplicate.slug, 'copy-of-data-2')
        self.assertEqual(data2_duplicate.name, 'Copy of Data 2')

        # Assert collection is not altered.
        self.assertEqual(Collection.objects.count(), 1)
        self.assertEqual(collection.entity_set.count(), 1)
        self.assertEqual(collection.entity_set.first().name, 'Entity')
        self.assertEqual(collection.data.count(), 2)
        collection_data = collection.data.all().order_by('name')
        self.assertEqual(collection_data[0].name, 'Data 1')
        self.assertEqual(collection_data[1].name, 'Data 2')

        # Assert permissions.
        self.assertEqual(len(get_perms(self.contributor, duplicate)), 6)

    def test_entity_duplicate_inherit(self):
        process = Process.objects.create(contributor=self.user)
        data = Data.objects.create(contributor=self.user, process=process, status=Data.STATUS_DONE)
        assign_perm('view_data', self.user, data)
        entity = Entity.objects.create(contributor=self.user)
        entity.data.add(data)

        # Add to collection.
        collection = Collection.objects.create(contributor=self.user)
        assign_perm('add_collection', self.user, collection)
        assign_perm('edit_collection', self.contributor, collection)
        entity.collections.add(collection.id)
        collection.data.add(data)

        # Duplicate.
        duplicated_entity1 = Entity.objects.filter(id=entity.id).duplicate(
            self.user,
            inherit_collections=False
        )[0]

        self.assertEqual(collection.entity_set.count(), 1)
        self.assertEqual(collection.entity_set.first().id, entity.id)
        self.assertEqual(collection.data.count(), 1)
        self.assertEqual(collection.data.first().id, data.id)

        # Assert permissions.
        self.assertEqual(duplicated_entity1.data.count(), 1)
        self.assertListEqual(get_perms(self.user, collection), ['add_collection'])
        self.assertEqual(len(get_perms(self.user, duplicated_entity1)), 6)
        self.assertEqual(len(get_perms(self.user, duplicated_entity1.data.first())), 5)
        self.assertListEqual(get_perms(self.contributor, collection), ['edit_collection'])
        self.assertListEqual(get_perms(self.contributor, duplicated_entity1), [])
        self.assertListEqual(get_perms(self.contributor, duplicated_entity1.data.first()), [])

        duplicated_entity2 = Entity.objects.filter(id=entity.id).duplicate(
            self.user,
            inherit_collections=True
        )[0]

        self.assertEqual(collection.entity_set.count(), 2)
        self.assertEqual(collection.entity_set.first().id, entity.id)
        self.assertEqual(collection.entity_set.last().id, duplicated_entity2.id)
        self.assertEqual(collection.data.count(), 2)
        self.assertEqual(collection.data.first().id, data.id)
        self.assertEqual(collection.data.last().id, duplicated_entity2.data.first().id)

        # Assert permissions.
        self.assertEqual(duplicated_entity2.data.count(), 1)
        self.assertListEqual(get_perms(self.user, collection), ['add_collection'])
        self.assertEqual(len(get_perms(self.user, duplicated_entity2)), 6)
        self.assertEqual(len(get_perms(self.user, duplicated_entity2.data.first())), 5)
        self.assertListEqual(get_perms(self.contributor, collection), ['edit_collection'])
        self.assertListEqual(get_perms(self.contributor, duplicated_entity2), ['edit_entity'])
        self.assertListEqual(get_perms(self.contributor, duplicated_entity2.data.first()), ['edit_data'])

    def test_collection_duplicate(self):
        entity = Entity.objects.create(name='Entity', contributor=self.user)
        assign_perm('view_entity', self.contributor, entity)

        # Add to collection.
        collection = Collection.objects.create(name='Collection', contributor=self.user)
        entity.collections.add(collection.id)

        # Duplicate.
        collections = Collection.objects.filter(id=collection.id).duplicate(self.contributor)
        self.assertEqual(len(collections), 1)
        duplicate = collections[0]

        collection_dict = Collection.objects.filter(id=collection.id).values()[0]
        duplicate_dict = Collection.objects.filter(id=duplicate.id).values()[0]

        entity_dict = Entity.objects.filter(id=entity.id).values()[0]
        entity_duplicate = duplicate.entity_set.first()
        entity_duplicate_dict = Entity.objects.filter(id=entity_duplicate.id).values()[0]

        self.assertTrue(duplicate.is_duplicate())
        self.assertTrue(entity_duplicate.is_duplicate())

        # Pop fields that should differ and assert the remaining.
        fields_to_differ = ('id', 'slug', 'contributor_id', 'name', 'duplicated', 'modified')
        for model_dict in (
                collection_dict, duplicate_dict,
                entity_dict, entity_duplicate_dict,
        ):
            for field in fields_to_differ:
                model_dict.pop(field)

        self.assertDictEqual(collection_dict, duplicate_dict)
        self.assertDictEqual(entity_dict, entity_duplicate_dict)

        # Assert fields that differ.
        self.assertEqual(duplicate.slug, 'copy-of-collection')
        self.assertEqual(duplicate.name, 'Copy of Collection')
        self.assertEqual(duplicate.contributor.username, 'contributor')
        self.assertAlmostEqual(duplicate.duplicated, datetime.now(), delta=timedelta(seconds=3))
        self.assertAlmostEqual(duplicate.modified, datetime.now(), delta=timedelta(seconds=3))

        self.assertEqual(entity_duplicate.slug, 'copy-of-entity')
        self.assertEqual(entity_duplicate.name, 'Copy of Entity')
        self.assertEqual(entity_duplicate.contributor.username, 'contributor')
        self.assertAlmostEqual(entity_duplicate.duplicated, datetime.now(), delta=timedelta(seconds=3))
        self.assertAlmostEqual(entity_duplicate.modified, datetime.now(), delta=timedelta(seconds=3))

        # Assert duplicated entity and data objects are in collection.
        self.assertEqual(duplicate.entity_set.count(), 1)
        self.assertEqual(duplicate.entity_set.first().name, 'Copy of Entity')

        # Assert permissions.
        self.assertEqual(len(get_perms(self.contributor, duplicate)), 6)


class ProcessModelTest(TestCase):

    def test_process_isactive(self):
        process = Process.objects.create(contributor=self.contributor)
        # is_active is true by default
        self.assertIs(process.is_active, True)
        process.is_active = False
        process.save()
        process_fetched = Process.objects.get(pk=process.pk)
        # the is_active flag is saved
        self.assertFalse(process_fetched.is_active)


@patch('resolwe.flow.models.utils.os')
class HydrateFileSizeUnitTest(TestCase):

    def create_data(self, os_mock, contributor, process):
        # Mock isfile and getsize for data creation to pass.
        # Make sure to set these two values to desired
        # values after calling this method.
        os_mock.path.isfile.return_value = True
        os_mock.path.getsize.return_value = 0

        data = Data.objects.create(contributor=contributor, process=process)
        data_location = DataLocation.objects.create(subpath='')
        data_location.subpath = str(data_location.id)
        data_location.save()
        data_location.data.add(data)

        data.output = {'test_file': {'file': 'test_file.tmp'}}
        data.save()

        return data

    def setUp(self):
        super().setUp()

        self.process = Process.objects.create(
            contributor=self.contributor,
            output_schema=[
                {'name': 'test_file', 'type': 'basic:file:', 'required': False},
                {'name': 'file_list', 'type': 'list:basic:file:', 'required': False}
            ]
        )

    def test_done_data(self, os_mock):
        data = self.create_data(os_mock, self.contributor, self.process)

        os_mock.path.isfile.return_value = True
        os_mock.path.getsize.return_value = 42000
        hydrate_size(data)
        self.assertEqual(data.output['test_file']['size'], 42000)

    def test_data_with_refs(self, os_mock):
        data = self.create_data(os_mock, self.contributor, self.process)

        os_mock.path.isfile.return_value = True
        os_mock.path.getsize.side_effect = [42000, 8000, 42]
        data.output = {
            'test_file': {
                'file': 'test_file.tmp',
                'refs': ['ref_file1.tmp', 'ref_file2.tmp']
            }
        }
        hydrate_size(data)
        self.assertEqual(data.output['test_file']['size'], 42000)
        self.assertEqual(data.output['test_file']['total_size'], 50042)
        self.assertEqual(data.size, 50042)

    def test_list(self, os_mock):
        data = self.create_data(os_mock, self.contributor, self.process)

        os_mock.path.isfile.return_value = True
        os_mock.path.getsize.side_effect = [34, 42000, 42]
        data.output = {
            'file_list': [
                {
                    'file': 'test_01.tmp',
                },
                {
                    'file': 'test_02.tmp',
                    'refs': ['ref_file1.tmp']
                },
            ]
        }
        hydrate_size(data)
        self.assertEqual(data.output['file_list'][0]['size'], 34)
        self.assertEqual(data.output['file_list'][0]['total_size'], 34)
        self.assertEqual(data.output['file_list'][1]['size'], 42000)
        self.assertEqual(data.output['file_list'][1]['total_size'], 42042)
        self.assertEqual(data.size, 34 + 42042)

    def test_change_size(self, os_mock):
        """Size is not changed after object is done."""
        data = self.create_data(os_mock, self.contributor, self.process)

        os_mock.path.isfile.return_value = True
        os_mock.path.getsize.return_value = 42000
        hydrate_size(data)
        self.assertEqual(data.output['test_file']['size'], 42000)

        os_mock.path.getsize.return_value = 43000
        hydrate_size(data)
        self.assertEqual(data.output['test_file']['size'], 43000)

        data.status = Data.STATUS_DONE
        os_mock.path.getsize.return_value = 44000
        hydrate_size(data)
        self.assertEqual(data.output['test_file']['size'], 43000)

    def test_missing_file(self, os_mock):
        data = self.create_data(os_mock, self.contributor, self.process)

        os_mock.path.isfile.return_value = False
        with self.assertRaises(ValidationError):
            hydrate_size(data)


class StorageModelTestCase(TestCase):

    def setUp(self):
        super().setUp()

        self.proc = Process.objects.create(
            name='Test process',
            contributor=self.contributor,
            output_schema=[
                {'name': 'json_field', 'type': 'basic:json:'},
            ],
        )

    def test_save_storage(self):
        """`basic:json:` fields are stored in Storage"""
        data = Data.objects.create(
            name='Test data',
            contributor=self.contributor,
            process=self.proc,
        )

        data.output = {'json_field': {'foo': 'bar'}}
        data.status = Data.STATUS_DONE
        data.save()

        self.assertEqual(Storage.objects.count(), 1)
        storage = Storage.objects.first()
        self.assertEqual(data.output['json_field'], storage.pk)

    def test_save_storage_file(self):
        """File is loaded and saved to storage"""
        data = Data.objects.create(
            name='Test data',
            contributor=self.contributor,
            process=self.proc,
        )
        data_location = DataLocation.objects.create(subpath='')
        data_location.subpath = str(data_location.id)
        data_location.save()
        data_location.data.add(data)

        data.output = {'json_field': 'json.txt'}
        data.status = Data.STATUS_DONE

        json_file = io.StringIO(json.dumps({'foo': 'bar'}))

        isfile_mock = MagicMock(return_value=True)
        open_mock = MagicMock(return_value=json_file)
        with patch.object(os.path, 'isfile', isfile_mock):
            with patch.object(builtins, 'open', open_mock):
                data.save()

        self.assertEqual(Storage.objects.count(), 1)
        storage = Storage.objects.first()
        self.assertEqual(data.output['json_field'], storage.pk)
        self.assertEqual(storage.json, {'foo': 'bar'})

    def test_delete_data(self):
        """Orphaned storages are deleted when data object is deleted"""
        storage_referenced_by_both = Storage.objects.create(contributor=self.user, json={})

        data_1 = Data.objects.create(contributor=self.contributor, process=self.proc)
        data_1.storages.create(contributor=self.user, json={})
        data_1.storages.add(storage_referenced_by_both)

        data_2 = Data.objects.create(contributor=self.contributor, process=self.proc)
        data_2.storages.add(storage_referenced_by_both)
        data_2.save()

        Storage.objects.create(contributor=self.user, json={})

        self.assertEqual(Data.objects.count(), 2)
        self.assertEqual(Storage.objects.count(), 3)

        # Delete first object.
        data_1.delete()
        self.assertEqual(Data.objects.count(), 1)
        self.assertEqual(Storage.objects.count(), 2)

        # Delete second object.
        data_2.delete()
        self.assertEqual(Data.objects.count(), 0)
        self.assertEqual(Storage.objects.count(), 1)

    def test_storage_manager(self):
        data = Data.objects.create(
            name='Test data',
            contributor=self.contributor,
            process=self.proc,
        )

        data.output = {'json_field': {'foo': {'moo': 'bar'}}}
        data.status = Data.STATUS_DONE
        data.save()

        # Annotation with specific JSON field.
        storage = Storage.objects.with_json_path(['foo', 'moo'])[0]
        self.assertEqual(storage.json_foo_moo, 'bar')

        storage = Storage.objects.with_json_path('foo.moo')[0]
        self.assertEqual(storage.json_foo_moo, 'bar')

        storage = Storage.objects.with_json_path(['foo', 'moo'], field='result')[0]
        self.assertEqual(storage.result, 'bar')

        # Extract specific JSON path.
        value = Storage.objects.get_json_path(['foo', 'moo'])[0]
        self.assertEqual(value, 'bar')

        value = Storage.objects.get_json_path('foo.moo')[0]
        self.assertEqual(value, 'bar')


class UtilsTestCase(TestCase):

    def test_render_template(self):
        process_mock = MagicMock(requirements={'expression-engine': 'jinja'})
        template = render_template(process_mock, '{{ 1 | increase }}', {})
        self.assertEqual(template, '2')

    def test_render_template_error(self):
        process_mock = MagicMock(requirements={'expression-engine': 'jinja'})
        with self.assertRaises(EvaluationError):
            render_template(process_mock, '{{ 1 | missing_increase }}', {})

    def test_hydrate_input_references(self):
        process = Process.objects.create(
            contributor=self.contributor,
            type='data:test:',
            output_schema=[
                {
                    'name': 'file',
                    'type': 'basic:file:',
                }, {
                    'name': 'file_list',
                    'type': 'list:basic:file:',
                }, {
                    'name': 'dir',
                    'type': 'basic:dir:',
                }, {
                    'name': 'dir_list',
                    'type': 'list:basic:dir:',
                },
            ],
        )
        descriptor_schema = DescriptorSchema.objects.create(
            contributor=self.contributor,
            schema=[
                {
                    'name': 'annotation',
                    'type': 'basic:string:',
                },
            ],
        )
        data = Data.objects.create(
            contributor=self.contributor,
            status=Data.STATUS_ERROR,
            process=process,
            # Workaround for skipping the validation.
            output={
                'file': {'file': 'some-file', 'refs': ['ref1']},
                'file_list': [{'file': 'some-file', 'refs': ['ref2']}, {'file': 'another-file'}],
                'dir': {'dir': 'some-dir', 'refs': ['ref3']},
                'dir_list': [{'dir': 'some-dir', 'refs': ['ref4']}, {'dir': 'another-dir'}],
            },
            descriptor_schema=descriptor_schema,
            descriptor={
                'annotation': 'my-annotation',
            },
            size=0,
        )
        data_location = DataLocation.objects.create(subpath='')
        data_location.subpath = str(data_location.id)
        data_location.save()
        data_location.data.add(data)

        input_schema = [
            {
                'name': 'data',
                'type': 'data:test:',
            },
        ]
        input_ = {'data': data.pk}
        hydrate_input_references(input_, input_schema)

        path_prefix = data.location.get_path()

        self.assertEqual(input_['data']['__descriptor'], {'annotation': 'my-annotation'})
        self.assertEqual(input_['data']['__type'], 'data:test:')
        self.assertEqual(input_['data']['__id'], data.id)

        self.assertEqual(input_['data']['file']['file'].data_id, data.id)
        self.assertEqual(input_['data']['file']['file'].file_name, 'some-file')
        self.assertEqual(str(input_['data']['file']['file']), os.path.join(path_prefix, 'some-file'))

        self.assertEqual(input_['data']['file']['refs'][0].data_id, data.id)
        self.assertEqual(input_['data']['file']['refs'][0].file_name, 'ref1')
        self.assertEqual(str(input_['data']['file']['refs'][0]), os.path.join(path_prefix, 'ref1'))

        self.assertEqual(input_['data']['file_list'][0]['file'].data_id, data.id)
        self.assertEqual(input_['data']['file_list'][0]['file'].file_name, 'some-file')
        self.assertEqual(str(input_['data']['file_list'][0]['file']), os.path.join(path_prefix, 'some-file'))

        self.assertEqual(input_['data']['file_list'][0]['refs'][0].data_id, data.id)
        self.assertEqual(input_['data']['file_list'][0]['refs'][0].file_name, 'ref2')
        self.assertEqual(str(input_['data']['file_list'][0]['refs'][0]), os.path.join(path_prefix, 'ref2'))

        self.assertEqual(input_['data']['file_list'][1]['file'].data_id, data.id)
        self.assertEqual(input_['data']['file_list'][1]['file'].file_name, 'another-file')
        self.assertEqual(str(input_['data']['file_list'][1]['file']), os.path.join(path_prefix, 'another-file'))

        self.assertEqual(input_['data']['dir']['dir'].data_id, data.id)
        self.assertEqual(input_['data']['dir']['dir'].file_name, 'some-dir')
        self.assertEqual(str(input_['data']['dir']['dir']), os.path.join(path_prefix, 'some-dir'))

        self.assertEqual(input_['data']['dir']['refs'][0].data_id, data.id)
        self.assertEqual(input_['data']['dir']['refs'][0].file_name, 'ref3')
        self.assertEqual(str(input_['data']['dir']['refs'][0]), os.path.join(path_prefix, 'ref3'))

        self.assertEqual(input_['data']['dir_list'][0]['dir'].data_id, data.id)
        self.assertEqual(input_['data']['dir_list'][0]['dir'].file_name, 'some-dir')
        self.assertEqual(str(input_['data']['dir_list'][0]['dir']), os.path.join(path_prefix, 'some-dir'))

        self.assertEqual(input_['data']['dir_list'][0]['refs'][0].data_id, data.id)
        self.assertEqual(input_['data']['dir_list'][0]['refs'][0].file_name, 'ref4')
        self.assertEqual(str(input_['data']['dir_list'][0]['refs'][0]), os.path.join(path_prefix, 'ref4'))

        self.assertEqual(input_['data']['dir_list'][1]['dir'].data_id, data.id)
        self.assertEqual(input_['data']['dir_list'][1]['dir'].file_name, 'another-dir')
        self.assertEqual(str(input_['data']['dir_list'][1]['dir']), os.path.join(path_prefix, 'another-dir'))



# pylint: disable=missing-docstring,invalid-name
import datetime

import mock

from django.apps import apps
from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser, Group
from django.core.management import call_command
from django.test import override_settings

from guardian.shortcuts import assign_perm
from rest_framework.test import APIRequestFactory, APITestCase, force_authenticate

from resolwe.elastic.builder import index_builder
from resolwe.test import TestCase

factory = APIRequestFactory()  # pylint: disable=invalid-name


CUSTOM_SETTINGS = {
    'INSTALLED_APPS': settings.INSTALLED_APPS + ('resolwe.elastic.tests.test_app',),
}


@override_settings(**CUSTOM_SETTINGS)
class IndexViewsetTest(APITestCase, TestCase):

    def setUp(self):
        from .test_app.models import TestModel
        from .test_app.elastic_indexes import TestSearchIndex
        from .test_app.viewsets import TestViewSet

        super().setUp()

        apps.clear_cache()
        call_command('migrate', verbosity=0, interactive=False, load_initial_data=False)

        index_builder.indexes = [TestSearchIndex()]
        index_builder.register_signals()

        # Prepare users and groups
        user_model = get_user_model()
        self.user_1 = user_model.objects.create(username='user_one')
        self.user_2 = user_model.objects.create(username='user_two')
        group = Group.objects.create(name='group')
        group.user_set.add(self.user_2)

        # Prepare test data
        test_obj_1 = TestModel.objects.create(name='Object name 1', number=43,
                                              date=datetime.datetime(2018, 1, 1, 0, 0))
        test_obj_2 = TestModel.objects.create(name='Object name 2', number=44,
                                              date=datetime.datetime(2017, 1, 1, 0, 0))
        test_obj_3 = TestModel.objects.create(name='Object name 3', number=45,
                                              date=datetime.datetime(2016, 1, 1, 0, 0))

        # Assing permissions
        assign_perm('view_testmodel', self.user_1, test_obj_1)
        assign_perm('view_testmodel', group, test_obj_2)
        assign_perm('view_testmodel', AnonymousUser(), test_obj_3)

        # Prepare test viewset
        self.test_viewset = TestViewSet.as_view(actions={
            'post': 'list_with_post',
        })

    def tearDown(self):
        index_builder.destroy()
        super().tearDown()

    def test_permissions(self):
        # First user
        request = factory.post('', {}, format='json')
        force_authenticate(request, self.user_1)
        response = self.test_viewset(request)
        response = sorted(response.data, key=lambda obj: obj['name'])

        self.assertEqual(len(response), 2)
        self.assertEqual(response[0]['name'], 'Object name 1')
        self.assertEqual(response[1]['name'], 'Object name 3')

        # Second user
        request = factory.post('', {}, format='json')
        force_authenticate(request, self.user_2)
        response = self.test_viewset(request)
        response = sorted(response.data, key=lambda obj: obj['name'])

        self.assertEqual(len(response), 2)
        self.assertEqual(response[0]['name'], 'Object name 2')
        self.assertEqual(response[1]['name'], 'Object name 3')

        # Public user
        request = factory.post('', {}, format='json')
        response = self.test_viewset(request)

        self.assertEqual(len(response.data), 1)
        self.assertEqual(response.data[0]['name'], 'Object name 3')

    def test_without_ordering(self):
        from .test_app.viewsets import TestEmptyOrderingViewSet

        viewset = TestEmptyOrderingViewSet.as_view(actions={
            'post': 'list_with_post',
        })

        request = factory.post('', {}, format='json')
        force_authenticate(request, self.user_1)
        response = viewset(request)
        response = sorted(response.data, key=lambda obj: obj['name'])

        self.assertEqual(len(response), 2)
        self.assertEqual(response[0]['name'], 'Object name 1')
        self.assertEqual(response[1]['name'], 'Object name 3')

    def test_pagination(self):
        request = factory.post('', {'offset': '0', 'limit': '1'}, format='json')
        force_authenticate(request, self.user_1)
        response = self.test_viewset(request)

        self.assertEqual(len(response.data['results']), 1)
        self.assertEqual(response.data['results'][0]['name'], 'Object name 1')

        request = factory.post('', {'offset': '1', 'limit': '1'}, format='json')
        force_authenticate(request, self.user_1)
        response = self.test_viewset(request)

        self.assertEqual(len(response.data['results']), 1)
        self.assertEqual(response.data['results'][0]['name'], 'Object name 3')

    @mock.patch('resolwe.elastic.viewsets.ELASTICSEARCH_SIZE', 1)
    def test_pagination_elasticsearch_size_limit(self):  # pylint: disable=invalid-name
        request = factory.post('', {'offset': '0', 'limit': '1'}, format='json')
        force_authenticate(request, self.user_1)
        response = self.test_viewset(request)

        self.assertEqual(len(response.data['results']), 1)

    def test_custom_filter(self):
        from .test_app.viewsets import TestCustomFieldFilterViewSet

        viewset = TestCustomFieldFilterViewSet.as_view(actions={
            'post': 'list_with_post',
        })

        request = factory.post('', {'name': '1'}, format='json')
        force_authenticate(request, self.user_1)
        response = viewset(request)

        self.assertEqual(len(response.data), 0)

        request = factory.post('', {'name': '43'}, format='json')
        force_authenticate(request, self.user_1)
        response = viewset(request)

        self.assertEqual(len(response.data), 1)
        self.assertEqual(response.data[0]['name'], 'Object name 1')

    def test_combined_viewset(self):
        from .test_app.viewsets import TestCombinedViewSet

        viewset = TestCombinedViewSet.as_view(actions={
            'get': 'list'
        })

        # Test database-only access.
        request = factory.get('', {}, format='json')
        force_authenticate(request, self.user_1)
        response = viewset(request)

        self.assertEqual(len(response.data), 2)
        self.assertEqual(response.data[0]['name'], 'Object name 1')
        self.assertEqual(response.data[0]['field_process_type'], '')
        self.assertEqual(response.data[0]['number'], 43)
        self.assertEqual(response.data[1]['name'], 'Object name 3')

        # Test combined access.
        request = factory.get('', {'name': '43'}, format='json')
        force_authenticate(request, self.user_1)
        response = viewset(request)

        self.assertEqual(len(response.data), 0)

        request = factory.get('', {'name': '1'}, format='json')
        force_authenticate(request, self.user_1)
        response = viewset(request)

        self.assertEqual(len(response.data), 1)
        self.assertEqual(response.data[0]['name'], 'Object name 1')
        self.assertEqual(response.data[0]['field_process_type'], '')
        self.assertEqual(response.data[0]['number'], 43)

        # Test combined order.
        request = factory.get('', {'name': 'Object', 'ordering': 'name'})
        force_authenticate(request, self.user_1)
        response = viewset(request)

        self.assertEqual(len(response.data), 2)
        self.assertEqual(response.data[0]['name'], 'Object name 1')
        self.assertEqual(response.data[1]['name'], 'Object name 3')

        request = factory.get('', {'name': 'Object', 'ordering': '-name'})
        force_authenticate(request, self.user_1)
        response = viewset(request)

        self.assertEqual(len(response.data), 2)
        self.assertEqual(response.data[0]['name'], 'Object name 3')
        self.assertEqual(response.data[1]['name'], 'Object name 1')

        # Test order by multiple fields.
        request = factory.get('', {'name': 'Object', 'ordering': '-name,num'})
        force_authenticate(request, self.user_1)
        response = viewset(request)

        self.assertEqual(len(response.data), 2)
        self.assertEqual(response.data[0]['name'], 'Object name 3')
        self.assertEqual(response.data[1]['name'], 'Object name 1')

    def _make_request(self, **kwargs):
        request = factory.post('', kwargs, format='json')
        force_authenticate(request, self.admin)
        return self.test_viewset(request)

    def test_filtering_map(self):
        response = self._make_request(name_alias='1')
        self.assertEqual(len(response.data), 1)

    def test_lookup_expressions_number_lt(self):
        response = self._make_request(num__lt='43')
        self.assertEqual(len(response.data), 0)

        response = self._make_request(num__lt='44')
        self.assertEqual(len(response.data), 1)

        response = self._make_request(num__lt='45')
        self.assertEqual(len(response.data), 2)

        response = self._make_request(num__lt='100')
        self.assertEqual(len(response.data), 3)

    def test_lookup_expressions_number_lte(self):
        response = self._make_request(num__lte='43')
        self.assertEqual(len(response.data), 1)

        response = self._make_request(num__lte='44')
        self.assertEqual(len(response.data), 2)

        response = self._make_request(num__lte='45')
        self.assertEqual(len(response.data), 3)

        response = self._make_request(num__lte='100')
        self.assertEqual(len(response.data), 3)

    def test_lookup_expressions_number_gt(self):
        response = self._make_request(num__gt='0')
        self.assertEqual(len(response.data), 3)

        response = self._make_request(num__gt='43')
        self.assertEqual(len(response.data), 2)

        response = self._make_request(num__gt='44')
        self.assertEqual(len(response.data), 1)

        response = self._make_request(num__gt='45')
        self.assertEqual(len(response.data), 0)

        response = self._make_request(num__gt='100')
        self.assertEqual(len(response.data), 0)

    def test_lookup_expressions_number_gte(self):
        response = self._make_request(num__gte='0')
        self.assertEqual(len(response.data), 3)

        response = self._make_request(num__gte='43')
        self.assertEqual(len(response.data), 3)

        response = self._make_request(num__gte='44')
        self.assertEqual(len(response.data), 2)

        response = self._make_request(num__gte='45')
        self.assertEqual(len(response.data), 1)

        response = self._make_request(num__gte='100')
        self.assertEqual(len(response.data), 0)

    def test_lookup_expressions_date_lt(self):
        response = self._make_request(date__lt='2018-03-22T12:23:14.430378')
        self.assertEqual(len(response.data), 3)

        response = self._make_request(date__lt='2017-03-22T12:23:14.430378')
        self.assertEqual(len(response.data), 2)

        response = self._make_request(date__lt='2016-03-22T12:23:14.430378')
        self.assertEqual(len(response.data), 1)

        response = self._make_request(date__lt='2018-01-01T00:00:00.000000')
        self.assertEqual(len(response.data), 2)

    def test_lookup_expressions_date_lte(self):
        response = self._make_request(date__lte='2018-03-22T12:23:14.430378')
        self.assertEqual(len(response.data), 3)

        response = self._make_request(date__lte='2017-03-22T12:23:14.430378')
        self.assertEqual(len(response.data), 2)

        response = self._make_request(date__lte='2016-03-22T12:23:14.430378')
        self.assertEqual(len(response.data), 1)

        response = self._make_request(date__lte='2018-01-01T00:00:00.000000')
        self.assertEqual(len(response.data), 3)

    def test_lookup_expressions_date_gt(self):
        response = self._make_request(date__gt='2018-03-22T12:23:14.430378')
        self.assertEqual(len(response.data), 0)

        response = self._make_request(date__gt='2017-03-22T12:23:14.430378')
        self.assertEqual(len(response.data), 1)

        response = self._make_request(date__gt='2016-03-22T12:23:14.430378')
        self.assertEqual(len(response.data), 2)

        response = self._make_request(date__gt='2018-01-01T00:00:00.000000')
        self.assertEqual(len(response.data), 0)

    def test_lookup_expressions_date_gte(self):
        response = self._make_request(date__gte='2018-03-22T12:23:14.430378')
        self.assertEqual(len(response.data), 0)

        response = self._make_request(date__gte='2017-03-22T12:23:14.430378')
        self.assertEqual(len(response.data), 1)

        response = self._make_request(date__gte='2016-03-22T12:23:14.430378')
        self.assertEqual(len(response.data), 2)

        response = self._make_request(date__gte='2018-01-01T00:00:00.000000')
        self.assertEqual(len(response.data), 1)

    def test_lookup_expressions_in(self):
        response = self._make_request(num__in='')
        self.assertEqual(len(response.data), 3)

        response = self._make_request(num__in='43')
        self.assertEqual(len(response.data), 1)
        self.assertEqual(response.data[0]['name'], 'Object name 1')

        response = self._make_request(num__in='45')
        self.assertEqual(len(response.data), 1)
        self.assertEqual(response.data[0]['name'], 'Object name 3')

        response = self._make_request(num__in='43,45')
        self.assertEqual(len(response.data), 2)
        self.assertEqual(response.data[0]['name'], 'Object name 1')
        self.assertEqual(response.data[1]['name'], 'Object name 3')

    def test_lookup_expressions_exact(self):
        from .test_app.models import TestModel

        TestModel.objects.create(name='Object name 4', number=45,
                                 date=datetime.datetime(2016, 1, 1, 0, 0),
                                 field_process_type='foo:bar:moo:')

        response = self._make_request(field_process_type='foo:bar')
        self.assertEqual(len(response.data), 1)
        self.assertEqual(response.data[0]['name'], 'Object name 4')

        response = self._make_request(field_process_type__exact='foo:bar')
        self.assertEqual(len(response.data), 0)

        response = self._make_request(field_process_type__exact='foo:bar:moo:')
        self.assertEqual(len(response.data), 1)
        self.assertEqual(response.data[0]['name'], 'Object name 4')



"""Python process runner utility."""
import argparse
import inspect
import json
import os
import sys
from importlib import import_module

from .runtime import Inputs, Process, ValidationError

if __name__ == '__main__':
    # pylint: disable=invalid-name
    parser = argparse.ArgumentParser(description="Run a Resolwe Python proces")
    parser.add_argument('filename', type=str,
                        help="Python process filename to run")
    parser.add_argument('--slug', type=str,
                        help="Slug of the process to run (required if multiple processes are defined)")
    parser.add_argument('--inputs', type=str,
                        help="Path to input JSON file")
    args = parser.parse_args()

    # Switch to target directory to import the module.
    try:
        filename = os.path.realpath(args.filename)
        start_dir = os.getcwd()
        os.chdir(os.path.dirname(filename))
        module, _ = os.path.splitext(os.path.basename(filename))
        module = import_module(module, __package__)
        os.chdir(start_dir)
    except (OSError, ImportError):
        print("ERROR: Failed to load Python process from '{}'.".format(args.filename))
        sys.exit(1)

    processes = {}
    for variable in dir(module):
        value = getattr(module, variable)
        if value == Process or not inspect.isclass(value) or not issubclass(value, Process):
            continue

        processes[value._meta.metadata.slug] = value  # pylint: disable=protected-access

    if args.slug:
        try:
            process = processes[args.slug]
        except KeyError:
            print("Found the following processes in module '{}':".format(args.module))
            print("")
            for slug, process in processes.items():
                print("  {} ({})".format(slug, process._meta.metadata.name))  # pylint: disable=protected-access
            print("")

            print("ERROR: Unable to find process '{}'.".format(args.slug))
            sys.exit(1)
    elif len(processes) == 1:
        process = next(iter(processes.values()))
    else:
        print("Found the following processes in module '{}':".format(args.module))
        print("")
        for slug, process in processes.items():
            print("  {} ({})".format(slug, process._meta.metadata.name))  # pylint: disable=protected-access
        print("")

        print("ERROR: Unable to determine which process to run. Pass --slug option.")
        sys.exit(1)

    # Prepare process inputs.
    inputs = Inputs()
    if args.inputs:
        with open(args.inputs) as inputs_file:
            data = json.load(inputs_file)

        for key, value in data.items():
            setattr(inputs, key, value)

    # TODO: Configure logging.

    # Start the process.
    instance = process()
    try:
        outputs = instance.start(inputs)
    except ValidationError as error:
        print("ERROR: Field validation failed: {}".format(error.args[0]))
        sys.exit(1)

    # Prepare outputs.
    for name, value in outputs.items():
        # TODO: Use the protocol to submit outputs.
        print(json.dumps(value))



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2017-12-04 16:07
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('flow', '0010_add_secret'),
    ]

    operations = [
        migrations.AlterField(
            model_name='datadependency',
            name='parent',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='children_dependency', to='flow.Data'),
        ),
    ]



"""Workflow execution engines."""
from resolwe.flow.engine import BaseEngine


class BaseExecutionEngine(BaseEngine):
    """A workflow execution engine."""

    def evaluate(self, data):
        """Return the code needed to compute a given Data object."""
        raise NotImplementedError

    def get_expression_engine(self, name):
        """Return an expression engine by its name."""
        return self.manager.get_expression_engine(name)

    def get_output_schema(self, process):
        """Return any additional output schema for the process."""
        return []

    def discover_process(self, path):
        """Perform process discovery in given path.

        This method will be called during process registration and
        should return a list of dictionaries with discovered process
        schemas.
        """
        return []

    def prepare_runtime(self, runtime_dir, data):
        """Prepare runtime directory.

        This method should return a dictionary of volume maps, where
        keys are files or directories relative the the runtime directory
        and values are paths under which these should be made available
        to the executing program. All volumes will be read-only.
        """
        pass
